
# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""Functions for performing parameter grid simulations with pyrates model_templates.
"""

# external imports
import pandas as pd
import numpy as np
from typing import Optional, Union, Tuple
from copy import deepcopy

# system imports
import os
import sys
import time as t
import glob
import getpass
import argparse
from pathlib import Path
from datetime import datetime
from threading import Thread, currentThread, RLock

# pyrates internal imports
from pyrates.frontend import CircuitTemplate
from pyrates.ir.circuit import CircuitIR

# meta infos
__author__ = "Christoph Salomon, Richard Gast"
__status__ = "development"


def grid_search(circuit_template: Union[CircuitTemplate, str], param_grid: Union[dict, pd.DataFrame], param_map: dict,
                dt: float, simulation_time: float, inputs: dict, outputs: dict,
                sampling_step_size: Optional[float] = None, permute_grid: bool = False, init_kwargs: dict = None,
                **kwargs) -> tuple:
    """Function that runs multiple parametrizations of the same circuit in parallel and returns a combined output.

    Parameters
    ----------
    circuit_template
        Path to the circuit template.
    param_grid
        Key-value pairs for each circuit parameter that should be altered over different circuit parametrizations.
    param_map
        Key-value pairs that map the keys of param_grid to concrete circuit variables.
    dt
        Simulation step-size in s.
    simulation_time
        Simulation time in s.
    inputs
        Inputs as provided to the `run` method of `:class:ComputeGraph`.
    outputs
        Outputs as provided to the `run` method of `:class:ComputeGraph`.
    sampling_step_size
        Sampling step-size as provided to the `run` method of `:class:ComputeGraph`.
    permute_grid
        If true, all combinations of the provided param_grid values will be realized. If false, the param_grid values
        will be traversed pairwise.
    kwargs
        Additional keyword arguments passed to the `:class:ComputeGraph` initialization.


    Returns
    -------
    tuple
        Simulation results stored in a multi-index data frame, the mapping between the data frame column names and the
        parameter grid, the simulation time, and the memory consumption.

    """

    # argument pre-processing
    #########################

    if not init_kwargs:
        init_kwargs = {}
    vectorization = init_kwargs.pop('vectorization', 'nodes')
    if type(circuit_template) is str:
        circuit_template = CircuitTemplate.from_yaml(circuit_template)

    # linearize parameter grid if necessary
    if type(param_grid) is dict:
        param_grid = linearize_grid(param_grid, permute_grid)

    # create grid-structure of network
    ##################################

    # get parameter names and grid length
    param_keys = list(param_grid.keys())
    N = param_grid.shape[0]

    # assign parameter updates to each circuit, combine them to unconnected network and remember their parameters
    circuit = CircuitIR()
    circuit_names = []
    for idx in param_grid.index:
        new_params = {}
        for key in param_keys:
            new_params[key] = param_grid[key][idx]
        circuit_tmp = deepcopy(circuit_template).apply()
        circuit_key = f'{circuit_tmp.label}_{idx}'
        circuit_tmp = adapt_circuit(circuit_tmp, new_params, param_map)
        circuit.add_circuit(circuit_key, circuit_tmp)
        circuit_names.append(circuit_key)
    param_grid.index = circuit_names

    # create backend graph
    net = circuit.compile(dt=dt, vectorization=vectorization, **init_kwargs)

    # adjust input of simulation to combined network
    for inp_key, inp in inputs.copy().items():
        inputs[f"all/{inp_key}"] = np.tile(inp, (1, N))
        inputs.pop(inp_key)

    # adjust output of simulation to combined network
    for out_key, out in outputs.items():
        outputs[out_key] = f"all/{out}"

    # simulate the circuits behavior
    results = net.run(simulation_time=simulation_time,
                      inputs=inputs,
                      outputs=outputs,
                      sampling_step_size=sampling_step_size,
                      **kwargs)    # type: pd.DataFrame

    # return results
    if 'profile' in kwargs:
        results, duration = results
        return results, param_grid, duration
    return results, param_grid


class ClusterCompute:
    def __init__(self, nodes: list, compute_dir=None, verbose: Optional[bool] = True):
        """Connect to nodes inside the computer network and create a compute directory with a unique compute ID

        Creates a compute directory for the `:class:ClusterCompute` instance in the specified path or as a default
        folder in the current working directory of the executing script.
        Creates a logfile in /ComputeDirectory/Logs and tees all future stdout and stderr of the executing script
        to this file
        Connects to all nodes via SSH and saves the corresponding paramiko client

        Parameters
        ----------
        nodes:
            List of names or IP addresses of working stations/servers in the local network
        compute_dir:
            Directory that will be used to store the logfiles.
            If none is provided, a default compute directory is created in the current working directory
        verbose:
            If False, all std output will still be copied to the log file but won't be shown in the terminal. Overwrites
            the verbosity of the run() method of all child classes.

        Returns
        -------

        """

        self.clients = []
        self.lock = RLock()

        #############################################################
        # Create main compute directory, subdirectories and logfile #
        #############################################################

        # Unique compute ID based on date and time
        self.compute_id = datetime.now().strftime("%d%m%y-%H%M%S")

        # Main compute directory
        if compute_dir:
            self.compute_dir = compute_dir
            os.makedirs(self.compute_dir, exist_ok=True)
        else:
            # Create default compute directory in the current working directory
            self.compute_dir = f'{os.getcwd()}/ClusterCompute_{self.compute_id}'
            os.makedirs(self.compute_dir, exist_ok=True)

        # Logfile directory
        self.log_dir = f'{self.compute_dir}/Logs/{self.compute_id}'
        os.makedirs(self.log_dir, exist_ok=True)

        # Global logfile to copy stdout and stderr to
        self.global_logfile = f'{self.log_dir}/Global_logfile.log'
        os.makedirs(os.path.dirname(self.global_logfile), exist_ok=True)

        ####################################
        # Tee stdout and stderr to logfile #
        ####################################
        if verbose:
            sys.stdout = StreamTee(sys.stdout, self.global_logfile)
            sys.stderr = StreamTee(sys.stderr, self.global_logfile)
        else:
            sys.stdout = open(self.global_logfile, 'a')
            sys.stderr = StreamTee(sys.stdout, self.global_logfile)

        ###########################
        # Write header to logfile #
        ###########################
        print("***NEW CLUSTER INSTANCE CREATED***")
        print(f'Compute ID: {self.compute_id}')
        print(f'Compute directory: {self.compute_dir}')
        print(f'Global logfile: {self.global_logfile}')
        print("")

        ##################
        # Create cluster #
        ##################
        print("***CONNECTING TO NODES...***")
        t0 = t.time()
        self.cluster_connect(nodes)
        print(f'Nodes connected. Elapsed time: {t.time()-t0:.3f} seconds')

    def __del__(self):
        # Make sure to close all clients when cluster_examples instance is destroyed
        for client in self.clients:
            try:
                client["paramiko_client"].close()
            except TypeError:
                pass

    def cluster_connect(self, nodes: list):
        """Create SSH connections to all nodes in the list, respectively

        Connect to all nodes in the given list via SSH, using the `ssh_connect` method of `:class:ClusterComoute`.
        Adds a dictionary for each node to the class internal 'clients' list
        Each dictionary contains:
            - ["paramiko_client"]: A paramiko client that can be used to execute commands on the node
            - ["node_name"]: The computer name of the connected node
            - ["hardware"]: A dictionary with certain hardware information of the node
            - ["logfile"]: Path to a logfile inside the instance's working directory,
                           where all stdout and stderr of the node will be redirected to

        Parameters
        ----------
        nodes
            List with names or IP addresses of working stations/servers in the local network

        Returns
        -------

        """
        username = getpass.getuser()
        # password = getpass.getpass(prompt='Enter password: ', stream=sys.stderr)
        for node in nodes:
            client = self.ssh_connect(node, username=username)
            if client is not None:
                # Create local logfiles
                local_logfile = f'{self.log_dir}/Local_logfile_{node}.log'
                os.makedirs(os.path.dirname(local_logfile), exist_ok=True)

                hw = self.get_hardware_spec(client)

                self.clients.append({
                    "paramiko_client": client,
                    "node_name": node,
                    "hardware": hw,
                    "logfile": local_logfile})

    def spawn_thread(self, client, thread_kwargs):
        """

        Parameters
        ----------
        client
        thread_kwargs

        Returns
        -------

        """
        t_ = Thread(
            name=client["node_name"],
            target=self.thread_master,
            args=(client, thread_kwargs)
        )
        t_.start()
        return t_

    def thread_master(self, client, thread_kwargs_: dict):
        """Function that is executed by every thread. Every instance of `thread_master` is bound to a different client

        The `thread_master` method of `:class:ClusterCompute` can be arbitrarily changed to fit the needs of the user.
        Commandline arguments on the remote node can be executed using client["paramiko_client"].exec_command("").

        Params
        ------
        client
            dict containing a `paramiko` client, the name of the connected node, a dict with hardware specifications and
            a path to a logfile.
        thread_kwargs

        Returns
        -------

        """
        pass

    @staticmethod
    def get_hardware_spec(pm_client):
        """Print hardware specifications of the workstation tied to the paramiko client

        Prints CPU model, number of CPU cores, min CPU freq, max CPU freq and working memory of a paramiko client
        connected to a remote worker node

        Parameters
        ----------
        pm_client
            Paramiko client

        Returns
        -------
        dict


        """

        try:
            stdin, stdout, stderr = pm_client.exec_command("lscpu | grep 'Model name'")
            cpu = stdout.readline().split()
            cpu = ' '.join(map(str, cpu[2:]))
        except IndexError:
            cpu = "N/A"
        print(f'CPU: {cpu}')

        try:
            stdin, stdout, stderr = pm_client.exec_command("lscpu | grep 'CPU(s)'")
            num_cpu_cores = int(stdout.readline().split(":")[1])
        except IndexError:
            num_cpu_cores = "N/A"
        print(f'Cores: {num_cpu_cores}')

        try:
            stdin, stdout, stderr = pm_client.exec_command("lscpu | grep 'max MHz'")
            cpu_freq = float(stdout.readline().split(":")[1])
        except IndexError:
            cpu_freq = "N/A"
        print(f'CPU freq: {cpu_freq} MHz')

        try:
            stdin, stdout, stderr = pm_client.exec_command("free -m | grep 'Mem'")
            total_mem = int(stdout.readline().split()[1])
        except IndexError:
            total_mem = "N/A"
        print(f'Total memory: {total_mem} MByte')

        print("")

        hw = {
            'cpu': cpu,
            'num_cpu_cores': num_cpu_cores,
            'cpu_freq': cpu_freq,
            'total_mem': total_mem
        }
        return hw

    @staticmethod
    def ssh_connect(node, username, password=None, print_status=True):
        """Connect to a host via SSH and return a respective `paramiko.SSHClient`

        Parameters
        ----------
        node
            Name or IP-address of the host to connect to
        username
        password

        Returns
        -------
        paramiko.SSHClient
            Is None if connection fails. For detailed information of the thrown exception see Paramiko documentation

        """
        import paramiko
        client = paramiko.SSHClient()
        try:
            # Using kerberos authentication
            client.connect(node, username=username, gss_auth=True, gss_kex=True)
            if print_status:
                print(f'\'{node}\': Connection established!')
            return client
        except paramiko.ssh_exception.NoValidConnectionsError as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None
        except paramiko.ssh_exception.AuthenticationException as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None
        except paramiko.ssh_exception.SSHException as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None
        except IOError as err:
            if print_status:
                print(f'\'{node}\': ', err)
            return None


class ClusterGridSearch(ClusterCompute):
    def __init__(self, nodes, compute_dir=None, verbose: Optional[bool] = True):
        """Connect to nodes inside the computer network and create a compute directory with a unique compute ID

        The use of `:class:ClusterGridSearch` requires the utilized computer network to feature two major properties:
        1. A GSSAPI-based authentication method (e.g. Kerberos).
           Since no secure password check is implemented in `:class:ClusterGridSearch`, it uses `paramiko's` gssapi
           module to authenticate users when an SSH connection is established with a worker.
           This has to be supported by the utilized computer network
        2. A server-based file system.
           After an SHH-connection has been established with a worker, all files that reside on the master have to be
           accessible by the worker using the same absolute file paths.
           A transfer of the necessary files from the master to the worker is generally possible using paramiko's SFTP
           module, but has not been implemented in `:class:ClusterGridSearch`.


        The compute directory contains the following sub folders:
        /Builds: Contains the build directories that are created by the `grid_search` function for each worker
        /Config: Contains configuration files in .json format that yield PyRates simulation parameters for each
                 simulation that has been performed by the `:class:ClusterGridSearch` instance.
        /Grids: Contains all parameter grids as .csv files that have been computed by the current CGS instance and their
                respective sub grids that were transferred to the workers as .h5 files.
        /Logs:  Contains the global log file and the local log files of each worker for each call to the `run` method of
                the current `:class:ClusterGridSearch` instance.
                The local log files contain the stdout and the stderr of each worker, respectively
                The global log files contains the stdout and the stderr of the master
        /Results: Contains all results in .h5 format that have been produced by the `:class:ClusterGridSearch` instance.
                  During a computation it also contains the intermediate local result files of the workers.
                  These files are deleted once the global result file created.

        Stdout and stderr of the worker will be teed to a global log file that resides in the compute directory.
        If the verbosity argument is set to False, all future stdout and stderr of the CGS instance will no longer be
        printed to the terminal, independent of the verbose argument of the call to the `run` method, but still be
        written to the global log file.

        Parameters
        ----------
        nodes
            List of names or IP addresses of working stations/servers in the local network
        compute_dir
            Directory that will be used to store the logfiles.
            If none is provided, a default compute directory is created in the current working directory
        verbose
            If False, all std output will still be copied to the log file but won't be shown in the terminal

        Returns
        -------

        """

        super().__init__(nodes, compute_dir, verbose)

        self.chunk_idx = 0
        self.res_file_collection = {}

        # Adding additional subfolders to the compute directory
        # Grid directory
        self.grid_dir = f'{self.compute_dir}/Grids'
        os.makedirs(self.grid_dir, exist_ok=True)

        # Config directory
        self.config_dir = f'{self.compute_dir}/Config'
        os.makedirs(self.config_dir, exist_ok=True)

        # Result directory
        self.res_dir = f'{self.compute_dir}/Results'
        os.makedirs(self.res_dir, exist_ok=True)

        # Build directory for pyrates backend
        self.build_dir = f'{self.compute_dir}/Builds'
        os.makedirs(self.build_dir, exist_ok=True)

    def run(self, circuit_template: str, param_grid: Union[dict, pd.DataFrame], param_map: dict, dt: float,
            simulation_time: float, chunk_size: (int, list),  inputs: dict, outputs: dict,
            worker_env: Optional[Union[str, list]] = sys.executable, worker_file: Optional[str] = os.path.abspath(__file__),
            sampling_step_size: Optional[float] = None, result_kwargs: Optional[dict] = {},
            worker_kwargs: Optional[dict] = {}, gs_kwargs: Optional[dict] = {},
            add_template_info: Optional[bool] = False, permute_grid: Optional[bool] = False,
            verbose: Optional[bool] = True, **kwargs) -> str:
        """Starts multiple threads on the master and distributes a parameter grid among nodes in a compute cluster that
        concurrently run a grid_search computation on their provided parameter sub grids.

        Intermediate result files will be concatenated to a single result file after all computations have finished.
        The global result file is a hierarchical data format (hdf5) file that contains the following groups and data
        sets.
        All groups that are marked with (df) can be loaded into a pandas.DataFrame, using pandas.read_hdf():

        /Config
            /circuit_template
            /config_file
            /dt
            /sampling_step_size
            /simulation_time
            /inputs (if provided)
        /ParameterGrid
            /Grid_df (df)
            /Keys
                /Parameter_1
                / ...
                /Parameter_p
        /Results
            /result_map (df)
            /results (df)
        /AdditionalData
            /result_kwargs_key_0
            /...
            /result_kwargs_key_k
        /TemplateInfo (if add_template_info is True)

        e.g. to access the stored simulation you can
        and performs Run multiple instances of grid_search simultaneously on different workstations in the compute cluster

        `:class:ClusterGridSearch` requires all necessary data (worker script, worker environment, compute directory)
        to be stored on a server. After a connection is established, all data on the remote workers has to be accessible
        in the same way as it would be on the master worker (same root directory, same filepaths, etc.)
        Paths inside the run function have to be adjusted respectively.

        FTP implementation is planned in future versions for direct file transfer between the master and the workers.


        Parameters
        ----------
        circuit_template
            Path to the circuit template file.
        param_grid
            Key-value pairs for each circuit parameter that should be altered over different circuit parametrizations.
        param_map
            Key-value pairs that map the keys of param_grid to circuit variables.
        dt
            Simulation step-size in s.
        simulation_time
            Simulation time in s.
        inputs
            inputs as provided to the `run` method of `:class:ComputeGraph`.
        outputs
            Outputs as provided to the `run` method of `:class:ComputeGraph`.
        sampling_step_size
            Sampling step-size as provided to the `run` method of `:class:ComputeGraph`.
        permute_grid
            If true, all combinations of the provided param_grid values will be realized. If false, the param_grid
            values will be traversed pairwise.
        chunk_size
            int or list that defines how many parametrizations are computed at once by each worker.
            If int, every worker fetches the specified amount of parametrizations. Worker repeatedly fetch parameter
            chunks until the whole param_grid has been computed.
            If list, ever worker fetches an individual amount of parametrizations during each computation.
            The order of chunk sizes in the chunk_size list is mapped to the order of nodes in the node list.
            Their lengths have to match.
        worker_env
            Path to python executable inside an environment that will be used to execute the worker file.
            Can be used if worker file with customized post processing should be executed in a different environment
            than the executing script.
        worker_file
            Path to customized worker file.
        result_kwargs
            Key-value pairs that will be added to the 'AdditionalData' group inside the result file.
        worker_kwargs
            Key-value pairs that will be added to the config file.
        gs_kwargs
            Key-Value pairs that will be passed to the grid_search() call of the workers
        add_template_info
            If True, all operator templates and its variables are copied from the circuit_template yaml file to the
            'TemplateInfo' group inside the result file.
        verbose
            If False, stdout and stderr will be copied to the log file but not be printed to the terminal

        Returns
        -------
        String containing location of the result file
        """

        import h5py

        if verbose:
            sys.stdout = StreamTee(sys.stdout, self.global_logfile)
            sys.stderr = StreamTee(sys.stderr, self.global_logfile)
        else:
            sys.stdout = open(self.global_logfile, 'a')
            sys.stderr = StreamTee(sys.stderr, self.global_logfile)

        t_total = t.time()

        print("")

        # Prepare parameter grid
        ########################
        print("***PREPARING PARAMETER GRID***")
        t0 = t.time()

        # Create DataFrame from param dictionary
        if isinstance(param_grid, dict):
            param_grid = linearize_grid(param_grid, permute=permute_grid)

        # Create default parameter grid csv-file
        grid_idx = 0
        while os.path.exists(f'{self.grid_dir}/DefaultGrid_{grid_idx}.csv'):
            grid_idx += 1
        grid_file = f'{self.grid_dir}/DefaultGrid_{grid_idx}.csv'
        grid_name = Path(grid_file).stem
        param_grid.to_csv(grid_file, index=True)

        # Add desired chunk size to each node
        #####################################
        for i, client in enumerate(self.clients):
            if isinstance(chunk_size, list):
                client['chunk_size'] = chunk_size[i]
            else:
                client['chunk_size'] = chunk_size

            if isinstance(worker_env, list):
                client['worker_env'] = worker_env[i]
            else:
                client['worker_env'] = worker_env

        # Assign chunk indices to parameter grid
        working_grid = param_grid.copy()
        working_grid["status"] = 'unsolved'
        working_grid["chunk_idx"] = -1
        working_grid["err_count"] = 0

        print(f'Done! Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Check key consistency
        #######################
        print("***CHECKING PARAMETER GRID AND MAP FOR CONSISTENCY***")
        t0 = t.time()
        if not self.check_key_consistency(param_grid, param_map):
            print("Terminating execution!")
            return ""
        print(f'Done! Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Create config file
        #####################
        print("***CREATING CONFIG FILE***")
        t0 = t.time()
        config_idx = 0
        while os.path.exists(f'{self.config_dir}/DefaultConfig_{config_idx}.json'):
            config_idx += 1
        config_file = f'{self.config_dir}/DefaultConfig_{config_idx}.yaml'
        self.create_cgs_config(fp=config_file,
                               circuit_template=circuit_template,
                               param_map=param_map,
                               dt=dt,
                               simulation_time=simulation_time,
                               inputs=inputs,
                               outputs=outputs,
                               sampling_step_size=sampling_step_size,
                               gs_kwargs=gs_kwargs,
                               worker_kwargs=worker_kwargs)
        print(f'Done! Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Create global result file
        ############################
        print("***CREATING GLOBAL RESULT FILE***")
        t0 = t.time()
        # Create result directory and result file for current parameter grid
        grid_res_dir = f'{self.res_dir}/{grid_name}'
        os.makedirs(grid_res_dir, exist_ok=True)
        global_res_file = f'{grid_res_dir}/CGS_result_{grid_name}.h5'

        # Write grid and config information to global result file
        with h5py.File(global_res_file, 'a') as file:
            for key, value in param_grid.items():
                file.create_dataset(f'/ParameterGrid/Keys/{key}', data=value)
            for key, value in result_kwargs.items():
                file.create_dataset(f'/AdditionalData/{key}', data=value)
            file.create_dataset(f'/Config/config_file', data=config_file)
            file.create_dataset(f'/Config/circuit_template', data=circuit_template)
            file.create_dataset(f'/Config/simulation_time', data=simulation_time)
            file.create_dataset(f'/Config/dt', data=dt)
            file.create_dataset(f'/Config/sampling_step_size', data=sampling_step_size)
            if add_template_info:
                template_file = f'{circuit_template.rsplit("/", 1)[0]}.yaml'
                self.add_template_information(template_file, file)

        param_grid.to_hdf(global_res_file, key='/ParameterGrid/Grid_df')
        print(f'Done. Elapsed time: {t.time() - t0:.3f} seconds')

        print("")

        # Create keyword dictionary for threads
        #######################################
        thread_kwargs = {
            "worker_env": worker_env,
            "worker_file": worker_file,
            "working_grid": working_grid,
            "grid_name": grid_name,
            "grid_res_dir": grid_res_dir,
            "config_file": config_file,
            "global_res_file": global_res_file
        }

        # Start cluster computation
        ###########################
        print("***STARTING CLUSTER COMPUTATION***")
        # Spawn threads to control each node connection and start computation
        threads = [self.spawn_thread(client, thread_kwargs) for client in self.clients]
        # Wait for all threads to finish
        for t_ in threads:
            t_.join()

        print("")
        print(f'Cluster computation finished. Elapsed time: {t.time() - t_total:.3f} seconds')

        # Check parameter combinations that failed multiple times to compute
        remaining_params = working_grid.loc[working_grid["status"] == "failed"].index
        if len(remaining_params) > 0:
            print("WARNING: The following parameter indices could not be computed: ")
            print(remaining_params)
        print("")

        # Write local results to global result file
        ###########################################
        print(f'***WRITING RESULTS TO GLOBAL RESULT FILE***')
        t0 = t.time()

        # Get sorted list of temporary result files to iterate through
        temp_res_files = glob.glob(f'{grid_res_dir}/*_temp*')
        temp_res_files.sort()

        # Read number of different circuit outputs and prepare lists to concatenate results
        res_dict = {}
        try:
            with pd.HDFStore(temp_res_files[0], "r") as store:
                for key in store.keys():
                    res_dict[key] = []
        except IndexError:
            print("No result file created. Check local log files for worker script errors.")
            return ""

        # Concatenate results from each temporary result file
        for file in temp_res_files:
            with pd.HDFStore(file, "r") as store:
                for idx, key in enumerate(store.keys()):
                    res_dict[key].append(store[key])

        # Create DataFrame for each output variable and write to global result file
        with pd.HDFStore(global_res_file, "a") as store:
            for key, value in res_dict.items():
                if key != '/result_map' and len(value) > 0:
                    df = pd.concat(value, axis=1)
                    store.put(key=f'/Results/{key}', value=df)
            result_map = pd.concat(res_dict['/result_map'], axis=0)
            store.put(key=f'/Results/result_map', value=result_map)

        # Delete temporary local result files
        for file in temp_res_files:
            os.remove(file)

        print(f'Elapsed time: {t.time()-t0:.3f} seconds')
        print(f'Find results in: {grid_res_dir}')
        print("")

        return global_res_file

    def thread_master(self, client: dict, thread_kwargs: dict):
        """Function that is executed by each thread to schedule computations on the respective worker

        Parameters
        ----------
        client
            Dictionary containing all information about the remote worker that is tied to the current thread
        thread_kwargs
            Dictionary containing all kwargs that are passed to the thread function

        Returns
        -------

        """

        import paramiko

        # This lock ensures that parameter chunks are fetched by workers in the same order as they are defined
        # in the node list
        self.lock.acquire()

        thread_name = currentThread().getName()
        connection_lost = False
        connection_lost_counter = 0

        # Get client information
        pm_client = client["paramiko_client"]
        logfile = client["logfile"]
        chunk_size = client["chunk_size"]
        worker_env = client["worker_env"]

        # Get keyword arguments
        config_file = thread_kwargs["config_file"]
        working_grid = thread_kwargs["working_grid"]
        grid_name = thread_kwargs["grid_name"]
        grid_res_dir = thread_kwargs["grid_res_dir"]
        worker_file = thread_kwargs["worker_file"]

        # Prepare worker command
        command = f'{worker_env} {worker_file}'

        # Create folder to save local subgrids to
        subgrid_dir = f'{self.grid_dir}/Subgrids/{grid_name}/{thread_name}'
        os.makedirs(subgrid_dir, exist_ok=True)
        subgrid_idx = 0

        # Create build_dir for grid_search numpy backend
        local_build_dir = f'{self.build_dir}/{thread_name}'
        os.makedirs(local_build_dir, exist_ok=True)

        # Start scheduler
        #################
        while True:
            # Disable thread switching
            self.lock.acquire()

            # Fetch grid chunks
            ####################
            # Get all parameters that haven't been successfully computed yet
            remaining_params = working_grid.loc[working_grid["status"] == "unsolved"]
            if remaining_params.empty:
                print(f'[T]\'{thread_name}\': No more parameter combinations available!')
                # Enable thread switching
                self.lock.release()
                # Release the initial lock in the first iteration
                try:
                    self.lock.release()
                except RuntimeError:
                    pass
                break

            else:
                # Find chunk index of first value in remaining params and fetch all parameter combinations with the
                # same chunk index
                param_idx, chunk_idx = self._fetch_index(remaining_params, chunk_size, working_grid)
                working_grid.loc[param_idx, "status"] = "pending"

                print(f'[T]\'{thread_name}\': Fetching {len(param_idx)} indices: ', end="")
                print(f'[{param_idx[0]}] - [{param_idx[-1]}]')

                # Create parameter sub-grid
                ###########################
                subgrid_fp = f'{subgrid_dir}/{thread_name}_Subgrid_{subgrid_idx}.h5'
                subgrid_df = working_grid.iloc[param_idx]
                subgrid_df.to_hdf(subgrid_fp, key="subgrid")
                subgrid_idx += 1

                # Temporary result file for current subgrid
                ###########################################
                local_res_file = f'{grid_res_dir}/CGS_result_{grid_name}_' \
                    f'chunk_{chunk_idx:02}_idx_{param_idx[0]}-{param_idx[-1]}_temp.h5'
                self.res_file_collection[chunk_idx] = local_res_file

                # Execute worker script on the remote host
                ##########################################
                t0 = t.time()

                try:
                    print(f'[T]\'{thread_name}\': Starting remote computation...')
                    stdin, stdout, stderr = pm_client.exec_command(command +
                                                                   f' --config_file={config_file}'
                                                                   f' --subgrid={subgrid_fp}'
                                                                   f' --local_res_file={local_res_file}'
                                                                   f' --build_dir={local_build_dir}'
                                                                   f' &>> {logfile}',  # redirect and append stdout
                                                                                       # and stderr to logfile
                                                                   get_pty=True)       # execute in pseudo terminal
                except paramiko.ssh_exception.SSHException as e:
                    # SSH connection has been lost
                    # (remote machine shut down, ssh connection has been killed manually, ...)
                    print(f'[T]\'{thread_name}\': ERROR: {e}')
                    working_grid.at[param_idx, "status"] = "unsolved"
                    connection_lost = True

            # Enable thread switching
            self.lock.release()

            # Release the second lock if its still acquired from the very beginning
            try:
                self.lock.release()
            except RuntimeError:
                pass

            # Try to reconnect to host if connection has been lost
            ######################################################
            if connection_lost and connection_lost_counter < 2:
                # Attempt to reconnect while there are still parameter chunks to fetch
                while True:
                    if working_grid.loc[working_grid["status"] == "unsolved"].empty:
                        # Stop thread execution if no more parameters are available
                        return
                    else:
                        print('Attempting to reconnect')
                        pm_client = self.ssh_connect(thread_name, username=getpass.getuser(), print_status=False)
                        if pm_client:
                            print(f'[T]\'{thread_name}\': Reconnected!')
                            connection_lost = False
                            # Escape reconnection loop
                            break
                        t.sleep(30)
                # Jump to the beginning of scheduler loop if reconnection was successful
                continue

            # Wait for remote computation exit status
            # (independent of success or failure)
            #########################################
            exit_status = stdout.channel.recv_exit_status()

            # Update grid status
            ####################
            with self.lock:
                # Remote machine executed worker script successfully
                if exit_status == 0:
                    print(f'[T]\'{thread_name}\': Remote computation finished. Elapsed time: {t.time()-t0:.3f} seconds')
                    print(f'[T]\'{thread_name}\': Updating grid status')
                    try:
                        # Check if data has been written to result file
                        with pd.HDFStore(local_res_file, "r") as store:
                            # Check if temporary result file (hdf5) contains keys
                            if len(store.keys()) == 0:
                                raise KeyError
                            # Check if each key contains data
                            for key in store.keys():
                                if len(store[key].index) == 0:
                                    raise KeyError
                        working_grid.at[param_idx, 'status'] = 'done'
                    except (KeyError, FileNotFoundError):
                        for row in param_idx:
                            if working_grid.loc[row]["err_count"] < 2:
                                working_grid.at[row, "status"] = "unsolved"
                                working_grid.at[row, "err_count"] = working_grid.iloc[row]["err_count"] + 1
                            else:
                                working_grid.at[row, "status"] = "failed"

                # Remote execution was interrupted
                else:
                    print(f'[T]\'{thread_name}\': Remote computation failed with exit status {exit_status}')
                    connection_lost_counter += 1
                    for row in param_idx:
                        if working_grid.loc[row]["err_count"] < 4:
                            working_grid.at[row, "status"] = "unsolved"
                            working_grid.at[row, "err_count"] = working_grid.iloc[row]["err_count"] + 1
                        else:
                            working_grid.at[row, "status"] = "failed"
                    if connection_lost_counter >= 2:
                        print('Excluding worker from pool')
                        return

            # Lock released, thread switching enabled
        # End of scheduler loop

        # TODO: Close pty on remote machine?

    # End of Thread master

    def _fetch_index(self, remaining_params, worker_chunk_size, working_grid):
        param_idx = []
        chunk_count = 0
        chunk_idx = self.chunk_idx
        for i, row in remaining_params.iterrows():
            if chunk_count == worker_chunk_size:
                self.chunk_idx += 1
                return pd.Index(param_idx), chunk_idx
            elif row['chunk_idx'] == -1:
                # Chunk parameter grid
                working_grid.loc[i, 'chunk_idx'] = self.chunk_idx
                param_idx.append(i)
                chunk_count += 1
            else:
                pass
        # Whole parameter grid is chunked
        if len(param_idx) == 0:
            # get all parameters with the same valid chunk index
            chunk_idx = remaining_params.iloc[0]["chunk_idx"]
            param_idx = remaining_params.loc[remaining_params["chunk_idx"] == chunk_idx].index

        return pd.Index(param_idx), chunk_idx

    #################################
    # CGS internal helper functions #
    #################################
    @staticmethod
    def chunk_grid(grid: pd.DataFrame, chunk_size: int):
        """Create a pd.DataFrame from the params dict and assign a chunk index to each parameter combination

        Parameters
        ----------
        grid
            pd.DataFrame with all parameter combinations that can be traversed linearly
        chunk_size
            Number of parameter combinations to be computed simultaneously on one worker

        Returns
        -------
            pd.DataFrame

        """
        # Add status columns to grid
        chunked_grid = grid.copy()
        chunked_grid["status"] = 'unsolved'
        chunked_grid["chunk_idx"] = -1
        chunked_grid["err_count"] = 0

        # Assign a chunk index to each parameter combination
        chunk_count = 0
        while True:
            param_idx = chunked_grid.loc[chunked_grid["status"] == "unsolved"].index[:chunk_size]
            if param_idx.empty:
                break
            else:
                chunked_grid.at[param_idx, 'chunk_idx'] = chunk_count
                chunked_grid.at[param_idx, 'status'] = "chunked"
                chunk_count += 1
        chunked_grid["status"] = 'unsolved'

        return chunked_grid

    @staticmethod
    def check_key_consistency(param_grid: Union[dict, pd.DataFrame], param_map: dict):
        """Check if keys in param_grid and param_map match

        Parameters
        ----------
        param_grid
        param_map

        Returns
        -------
        bool

        """
        grid_key_lst = list(param_grid.keys())
        map_key_lst = list(param_map.keys())
        # Not all keys of parameter map can be found in parameter grid
        if not all((grid_key in map_key_lst for grid_key in grid_key_lst)):
            print("Not all parameter grid keys found in parameter map")
            print("Parameter map keys:")
            print(*list(param_map.keys()))
            print("Parameter grid keys:")
            print(*list(param_grid.keys()))
            return False
        # Parameter grid and parameter map match
        else:
            print("All parameter grid keys found in parameter map")
            return True

    @staticmethod
    def create_cgs_config(fp: str, circuit_template: str, param_map: dict, dt: float, simulation_time: float,
                          sampling_step_size: Optional[float], inputs: Optional[dict] = {},
                          outputs: Optional[dict] = {}, gs_kwargs: Optional[dict] = {}, worker_kwargs: Optional[dict] = {}):
        """Creates a configfile.json containing a dictionary with all input parameters as key-value pairs

        Parameters
        ----------
        fp
        circuit_template
        param_map
        dt
        simulation_time
        inputs
        outputs
        sampling_step_size
        gs_kwargs
        worker_kwargs

        Returns
        -------

        """
        
        import yaml
        
        if not sampling_step_size:
            sampling_step_size = dt

        config_dict = {
            "circuit_template": circuit_template,
            "param_map": param_map,
            "dt": dt,
            "simulation_time": simulation_time,
            "inputs": inputs,
            "outputs": outputs,
            "sampling_step_size": sampling_step_size,
            "gs_kwargs": gs_kwargs,
            "worker_kwargs": worker_kwargs
        }

        with open(fp, "w") as f:
            yaml.dump(config_dict, f, default_flow_style=False, sort_keys=False)

    @staticmethod
    def add_template_information(yaml_fp, hdf5_file):
        """Add opearator information of the circuit template to the global result file"""
        
        import yaml
        
        with open(yaml_fp, 'r') as stream:
            try:
                yaml_dict = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)

        for operator_key, operator_value in yaml_dict.items():
            if "Op" in operator_key:
                for temp_key, temp_value in operator_value.items():
                    if isinstance(temp_value, str):
                        hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}', data=temp_value)
                    if isinstance(temp_value, list):
                        for idx, eq in enumerate(temp_value):
                            hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}/eq_{idx}', data=eq)
                    elif isinstance(temp_value, dict):
                        for key, value in temp_value.items():
                            try:
                                hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}/{key}', data=value["default"])
                            except:
                                hdf5_file.create_dataset(f'/TemplateInfo/{operator_key}/{temp_key}/{key}', data=value)

#####################
# Utility functions #
#####################


def linearize_grid(grid: dict, permute: bool = False) -> pd.DataFrame:
    """Turns the grid into a grid that can be traversed linearly, i.e. pairwise.

    Parameters
    ----------
    grid
        Parameter grid.
    permute
        If true, all combinations of the parameter values in grid will be created.

    Returns
    -------
    pd.DataFrame
        Resulting linear grid in form of a data frame.

    """

    arg_lengths = [len(arg) for arg in grid.values()]

    if len(list(set(arg_lengths))) == 1 and not permute:
        return pd.DataFrame(grid)
    else:
        vals, keys = [], []
        for key, val in grid.items():
            vals.append(val)
            keys.append(key)
        new_grid = np.stack(np.meshgrid(*tuple(vals)), -1).reshape(-1, len(grid))
        return pd.DataFrame(new_grid, columns=keys)


def adapt_circuit(circuit: CircuitIR, params: dict, param_map: dict) -> CircuitIR:
    """Changes the parametrization of a circuit.

    Parameters
    ----------
    circuit
        Circuit instance.
    params
        Key-value pairs of the parameters that should be changed.
    param_map
        Map between the keys in params and the circuit variables.

    Returns
    -------
    CircuitIR
        Updated circuit instance.

    """

    for key in params.keys():

        val = params[key]

        for var in param_map[key]['vars']:

            # change variable values on nodes
            nodes = param_map[key]['nodes'] if 'nodes' in param_map[key] else []
            for node in nodes:
                if "/" in var:
                    op, var_name = var.split("/")
                    if op in circuit[node]:
                        circuit[node].values[op][var_name] = float(val)
                else:
                    for op, _ in circuit[node]:
                        try:
                            circuit[node].values[op][var] = float(val)
                        except KeyError:
                            pass

            # change variable values on edges
            edges = param_map[key]['edges'] if 'edges' in param_map[key] else []
            if edges and len(edges[0]) < 3:
                for source, target in edges:
                    if var in circuit.edges[source, target, 0]:
                        circuit.edges[source, target, 0][var] = float(val)
            else:
                for source, target, edge in edges:
                    if var in circuit.edges[source, target, edge]:
                        circuit.edges[source, target, edge][var] = float(val)

    return circuit


class StreamTee(object):
    """Tee all stdout to a specified logfile"""
    def __init__(self, stream1, stream2fp):
        stream2 = open(stream2fp, "a")
        self.stream1 = stream1
        self.stream2 = stream2
        self.__missing_method_name = None  # Hack!

    def __getattribute__(self, name):
        return object.__getattribute__(self, name)

    def __getattr__(self, name):
        self.__missing_method_name = name  # Could also be a property
        return getattr(self, '__methodmissing__')

    def __methodmissing__(self, *args, **kwargs):
        # Emit method call to the log copy
        callable2 = getattr(self.stream2, self.__missing_method_name)
        callable2(*args, **kwargs)

        # Emit method call to stdout (stream 1)
        callable1 = getattr(self.stream1, self.__missing_method_name)
        return callable1(*args, **kwargs)


##################
# Cluster worker #
##################


class ClusterWorkerTemplate(object):
    """Contains an interface for its methods to be called on a cluster worker by a remote master.

    """
    def __init__(self):
        self.FLAGS = argparse.Namespace
        self.results = pd.DataFrame
        self.result_map = pd.DataFrame
        self.processed_results = pd.DataFrame

    def worker_init(self, config_file="", subgrid="", result_file="", build_dir=os.getcwd()):
        """Interface to receive input when run from the console

        Returns
        -------

        """
        parser = argparse.ArgumentParser()

        parser.add_argument(
            "--config_file",
            type=str,
            default=config_file,
            help="File to load grid_search configuration parameters from"
        )

        parser.add_argument(
            "--subgrid",
            type=str,
            default=subgrid,
            help="File to load parameter grid from"
        )

        parser.add_argument(
            "--local_res_file",
            type=str,
            default=result_file,
            help="File to save results to"
        )

        parser.add_argument(
            "--build_dir",
            type=str,
            default=build_dir,
            help="Custom PyRates build directory"
        )

        self.FLAGS = parser.parse_args()
        self.worker_exec(sys.argv)

    def worker_exec(self, _):
        """Reads configurations and sub grid from the respective files and executes a grid_search() call with the
           provided parameters.

        Parameters
        ----------
        _

        Returns
        -------

        """
        import warnings
        # external imports
        from numba import config
        import yaml

        # tf.config.set_soft_device_placement(True)

        config.THREADING_LAYER = 'omp'

        # Disable general warnings
        warnings.filterwarnings("ignore")

        # disable TF-gpu warnings
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

        t_total = t.time()

        # Load command line arguments and create logfile
        ################################################
        print("")
        print("***LOADING COMMAND LINE ARGUMENTS***")
        t0 = t.time()

        config_file = self.FLAGS.config_file
        subgrid = self.FLAGS.subgrid
        local_res_file = self.FLAGS.local_res_file
        build_dir = self.FLAGS.build_dir

        print(f'Elapsed time: {t.time() - t0:.3f} seconds')

        # Load global config file
        #########################
        print("")
        print("***LOADING GLOBAL CONFIG FILE***")
        t0 = t.time()

        with open(config_file, "r") as g_conf:
            global_config_dict = yaml.load(stream=g_conf, Loader=yaml.UnsafeLoader)

        circuit_template = global_config_dict['circuit_template']
        param_map = global_config_dict['param_map']
        dt = global_config_dict['dt']
        simulation_time = global_config_dict['simulation_time']

        # Optional parameters
        #####################
        try:
            sampling_step_size = global_config_dict['sampling_step_size']
        except KeyError:
            sampling_step_size = dt

        try:
            inputs = global_config_dict['inputs']
        except KeyError:
            inputs = {}

        try:
            outputs = global_config_dict['outputs']
        except KeyError:
            outputs = {}

        try:
            worker_kwargs = global_config_dict['worker_kwargs']
        except KeyError:
            worker_kwargs = {}

        try:
            gs_kwargs = global_config_dict['gs_kwargs']
        except KeyError:
            gs_kwargs = {}

        print(f'Elapsed time: {t.time() - t0:.3f} seconds')

        # LOAD PARAMETER GRID
        #####################
        print("")
        print("***PREPARING PARAMETER GRID***")
        t0 = t.time()

        # Load subgrid into DataFrame
        param_grid = pd.read_hdf(subgrid, key="subgrid")

        # Drop all columns that don't contain a parameter map value (e.g. status, chunk_idx, err_count) since
        # grid_search() can't handle additional columns
        param_grid = param_grid[list(param_map.keys())]
        print(f'Elapsed time: {t.time() - t0:.3f} seconds')

        # COMPUTE PARAMETER GRID
        ########################
        print("")
        print("***COMPUTING PARAMETER GRID***")
        t0 = t.time()

        self.results, self.result_map, t_ = grid_search(
            circuit_template=circuit_template,
            param_grid=param_grid,
            param_map=param_map,
            simulation_time=simulation_time,
            dt=dt,
            sampling_step_size=sampling_step_size,
            permute_grid=False,
            inputs=inputs,
            outputs=outputs.copy(),
            profile="t",
            build_dir=build_dir,
            **gs_kwargs)

        print(f'Total parameter grid computation time: {t.time() - t0:.3f} seconds')

        # Post process results and write data to local result file
        ##########################################################
        print("")
        print("***POSTPROCESSING AND CREATING RESULT FILES***")
        t0 = t.time()

        self.processed_results = pd.DataFrame(data=None, columns=self.results.columns)
        self.worker_postprocessing(**worker_kwargs)

        with pd.HDFStore(local_res_file, "w") as store:
            store.put(key='results', value=self.processed_results)
            store.put(key='result_map', value=self.result_map)

        print(f'Result files created. Elapsed time: {t.time() - t0:.3f} seconds')
        print("")
        print(f'Total elapsed time: {t.time() - t_total:.3f} seconds')

    def worker_postprocessing(self, **worker_kwargs):
        """
        Post processing that is applied by each cluster worker on its computed model output.

        The ClusterWorkerTemplate class contains three DataFrames that can be used to add customized post processing:

        To apply the customized post processing during a ClusterGridSearch call, a customized worker file has to
        be created that is passed to the CGS.run() call.
        For that purpose, execute the following steps:
        1. Create a new python script.
        2. Derive a child class from the ClusterWorkerTemplate class and customize its worker_postprocessing() method.
           Use self.results and self.result_map to access simulation data.
           Safe the post processed data to self.processed_results.

           self.results - DataFrame that contains the worker's grid_search() output for the provided parameter sub grid
           self.result_map - DataFrame that contains the computed parametrizations and their respective identification
                             key which is used in the column names of the results DataFrame
           self.processed_results - DataFrame that will be send back to the master.
                                    Has the same column structure as self.results, yet the index is unset.
                                    The processed_results of all workers will be concatenated to a the final result

        3. Ensure that if the worker script is executed, an instance of your customized worker class is invoked and that
           its 'worker_init()' method is called.

        When implementing customized post processing, use the following code to iterate over all computed results.
        Make sure to write all final data to self.processed_results:

        for idx, data in self.results.iteritems():

            # add processing of 'data' here

            self.processed_results[:, idx] = data

        An example worker template that estimates the power spectral density of each model parametrization results is
        presented below:

            custom_worker.py

            from pyrates.utility.grid_search import ClusterWorkerTemplate
            from scipy.signal import welch


            class MyWorker(ClusterWorkerTemplate):
                def worker_postprocessing(self):
                    for idx, data in self.results.iteritems():
                        t = self.results.index.to_list()
                        dt = t[1] - t[0]
                        f, p = welch(data.to_numpy(), fs=1/dt, axis=0)
                        self.processed_results[:, idx] = p
                    self.processed_results.index = f


            if __name__ == "__main__":
                my_worker = MyWorker()
                my_worker.worker_init()

        Parameters
        ----------
        kwargs

        Returns
        -------

        """

        for idx, data in self.results.iteritems():

            # Add post customized post processing of 'data' here

            self.processed_results.loc[:, idx] = data

    def worker_test(self):
        """
        Calls the worker_postprocessing script on a dummy result and a dummy result map

        self.result_map dummy looks as follows:

                       param_0  param_1
            circuit_0        0        9
            circuit_1        1        8
            circuit_2        2        7
            circuit_3        3        6
            circuit_4        4        5
            circuit_5        5        4
            circuit_6        6        3
            circuit_7        7        2
            circuit_8        8        1
            circuit_9        9        0

        self.results dummy looks as follows:

          circuit_0 circuit_1 circuit_2  ... circuit_7 circuit_8 circuit_9
                pop       pop       pop  ...       pop       pop       pop
            out_var   out_var   out_var  ...   out_var   out_var   out_var
        0       0.0       1.0       2.0  ...       7.0       8.0       9.0
        1      10.0      11.0      12.0  ...      17.0      18.0      19.0
        2      20.0      21.0      22.0  ...      27.0      28.0      29.0
        3      30.0      31.0      32.0  ...      37.0      38.0      39.0
        4      40.0      41.0      42.0  ...      47.0      48.0      49.0
        5      50.0      51.0      52.0  ...      57.0      58.0      59.0
        6      60.0      61.0      62.0  ...      67.0      68.0      69.0
        7      70.0      71.0      72.0  ...      77.0      78.0      79.0
        8      80.0      81.0      82.0  ...      87.0      88.0      89.0
        9      90.0      91.0      92.0  ...      97.0      98.0      99.0

        Returns
        -------

        """
        # Create Dummies
        circuit_names = [f'circuit_{i}' for i in range(10)]

        grid = {'param_0': list(range(10)),
                'param_1': list(range(10))[::-1]}

        dummy_grid = linearize_grid(grid, permute=False)
        dummy_grid.index = circuit_names

        pop = ['pop'] * 10
        out_var = ['out_var'] * 10

        dummy_mi = pd.MultiIndex.from_arrays([circuit_names, pop, out_var])
        dummy_results = pd.DataFrame(np.arange(100.0).reshape((10, 10)), columns=dummy_mi)

        self.result_map = dummy_grid
        self.results = dummy_results
        self.processed_results = pd.DataFrame(data=None, columns=self.results.columns)

        # Run post processing method
        self.worker_postprocessing()

        print('Test successful, no errors occurred!')


if __name__ == "__main__":
    cgs_worker = ClusterWorkerTemplate()
    # cgs_worker.worker_test()
    cgs_worker.worker_init()






# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation

# external imports
import pandas as pd
import numpy as np
from typing import List

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


def functional_connectivity(data: pd.DataFrame, metric: str = 'cov', **kwargs) -> np.ndarray:
    """Calculate functional connectivity of node timeseries in data.

    Parameters
    ----------
    data
        Pandas dataframe containing the simulation results.
    metric
        Type of connectivtiy measurement that should be used.
            - `cov` for covariance (uses `np.cov`)
            - `corr` for pearsson correlation (uses `np.corrcoef`)
            - `csd` for cross-spectral density (uses `mne.time_frequency.csd_array_morlet`)
            - `coh` for coherenc (uses `mne.connectivtiy.spectral_connectivity`)
            - `cohy` for coherency (uses `mne.connectivtiy.spectral_connectivity`)
            - `imcoh` for imaginary coherence (uses `mne.connectivtiy.spectral_connectivity`)
            - `plv` for phase locking value (uses `mne.connectivtiy.spectral_connectivity`)
            - `ppc` for pairwise phase consistency (uses `mne.connectivtiy.spectral_connectivity`)
            - `pli` for phase lag index (uses `mne.connectivtiy.spectral_connectivity`)
            - `pli2_unbiased` for unbiased estimate of squared phase lag index
               (uses `mne.connectivtiy.spectral_connectivity`)
            - `wpli`for weighted phase lag index (uses `mne.connectivtiy.spectral_connectivity`)
            - `wpli2_debiased` for debiased weighted phase lag index (uses `mne.connectivtiy.spectral_connectivity`)
    kwargs
        Additional keyword arguments passed to respective function used for fc calculation.

    Returns
    -------
    np.ndarray
        Pairwise functional connectivity

    """

    if 'time' in data.columns.values:
        idx = data.pop('time')
        data.index = idx

    # calculate functional connectivity
    ###################################

    if metric == 'cov':

        # covariance
        fc = np.cov(data.values.T, **kwargs)

    elif metric == 'corr':

        # pearsson correlation coefficient
        fc = np.corrcoef(data.values.T, **kwargs)

    elif metric == 'csd':

        from mne.time_frequency import csd_array_morlet
        fc = np.abs(csd_array_morlet(X=np.reshape(data.values, (1, data.shape[1], data.shape[0])),
                                     sfreq=1./(data.index[1] - data.index[0]),
                                     ch_names=data.columns.values,
                                     **kwargs).mean().get_data())

    elif metric in 'cohcohyimcohplvppcplipli2_unbiasedwpliwpli2_debiased':

        # phase-based connectivtiy/synchronization measurement
        from mne.connectivity import spectral_connectivity
        fc, _, _, _, _ = spectral_connectivity(np.reshape(data.values.T, (1, data.shape[1], data.shape[0])),
                                               method=metric,
                                               sfreq=1./(data.index[1] - data.index[0]),
                                               **kwargs)
        fc = fc[:, :, 0]

    else:

        raise ValueError(f'FC metric is not supported by this function: {metric}. Check the documentation of the '
                         f'argument `metric` for valid options.')

    return fc


def analytic_signal(data: pd.DataFrame, fmin: float, fmax: float, nodes: List[str] = None, **kwargs) -> pd.DataFrame:
    """Calculates analytic signal from simulation results, using the hilbert transform.

    Parameters
    ----------
    data
        Simulation results.
    fmin
        Lower bound frequency for bandpass filter that will be applied to the data.
    fmax
        Upper bound frequency for bandpass filter that will be applied to the data.
    nodes
        List of node names for which to calculate the analytic signal.
    kwargs
        Additional keyword arguments that will be passed to the `mne.Raw.filter` method.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the fields `time`, `node`, `amplitude` and `phase`.

    """

    if 'time' in data.columns.values:
        idx = data.pop('time')
        data.index = idx

    if nodes:
        if type(nodes[0]) is str:
            data = data.loc[:, nodes]
        else:
            data = data.iloc[:, nodes]

    # create mne raw data object
    from pyrates.utility import mne_from_dataframe
    raw = mne_from_dataframe(data)

    # bandpass filter the raw data
    raw.filter(l_freq=fmin, h_freq=fmax, **kwargs)

    # apply hilbert transform
    raw.apply_hilbert()

    # get phase of analytic signal
    def get_angle(x):
        return np.angle(x) + np.pi
    raw_phase = raw.copy()
    raw_phase.apply_function(get_angle)
    raw_phase.apply_function(np.real, dtype=np.float32)
    raw_phase.apply_function(np.unwrap)

    # get amplitude of analytic signal
    raw_amplitude = raw.copy()
    raw_amplitude.apply_function(np.abs)
    raw_amplitude.apply_function(np.real, dtype=np.float32)

    # combine phase and amplitude into dataframe
    time = data.index
    data_phase = raw_phase.to_data_frame(scalings={'eeg': 1.})
    data_phase['time'] = time
    data_amp = raw_amplitude.to_data_frame(scalings={'eeg': 1.})
    data_amp['time'] = time
    data = pd.melt(data_phase, id_vars=['time'], var_name='node', value_name='phase')
    data_tmp = pd.melt(data_amp, id_vars=['time'], var_name='node', value_name='amplitude')
    data['amplitude'] = data_tmp['amplitude']

    return data


def welch(data, tmin=0., tmax=None, **kwargs):
    """

    Parameters
    ----------
    data
    tmin
    tmax
    kwargs

    Returns
    -------

    """

    # prepare data frame
    dt = data.index[1] - data.index[0]
    tmin = int(tmin / dt)
    tmax = data.shape[0] + 1 if tmax is None else max([int(tmax/dt), data.shape[0] + 1])
    if len(data.shape) > 1:
        data = data.iloc[tmin:tmax, :]
    else:
        data = data.iloc[tmin:tmax]

    # Compute power spectral density
    try:
        from scipy.signal import welch
        return welch(data.values, fs=1./dt, axis=0, **kwargs)
    except IndexError:
        return np.NaN, np.NaN


def fft(data, tmin=0., **kwargs):
    """

    Parameters
    ----------
    data
    tmin

    Returns
    -------

    """
    # Compute spectrum
    try:
        dt = data.index[1] - data.index[0]
        n = data.shape[0]

        # Get closest power of 2 that includes n for zero padding
        n_two = 1 if n == 0 else 2 ** (n - 1).bit_length()

        data_tmp = data.loc[tmin:, :]
        data_tmp = data_tmp - np.mean(data_tmp)

        freqs = np.linspace(0, 1 / dt, n_two)
        spec = np.fft.fft(data_tmp, n=n_two, axis=0, **kwargs)

        # Cut of PSD and frequency arrays since its mirrored at N/2
        spec = np.abs(spec[:int(len(spec) / 2)])
        freqs = freqs[:int(len(freqs) / 2)]
        return freqs, spec
    except IndexError:
        return np.NaN, np.NaN


def time_frequency(data: pd.DataFrame, freqs: List[float], method: str = 'morlet', output: str = 'avg_power', **kwargs
                   ) -> np.ndarray:
    """Calculates time-frequency representation for each node.

    Parameters
    ----------
    data
        Simulation results.
    freqs
        Frequencies of interest.
    method
        Method to be used for TFR calculation. Can be `morlet` for `mne.time_frequency.tfr_array_morlet` or
        `multitaper` for `mne.time_frequency.tfr_array_multitaper`.
    output
        Type of the output variable to be calculated. For options, see `mne.time_frequency.tfr_array_morlet`.
    kwargs
        Additional keyword arguments to be passed to the function used for tfr calculation.

    Returns
    -------
    np.ndarray
        Time-frequency representation (n x f x t) for each node (n) at each frequency of interest (f) and time (t).

    """

    if 'time' in data.columns.values:
        idx = data.pop('time')
        data.index = idx

    if method == 'morlet':

        from mne.time_frequency import tfr_array_morlet
        return tfr_array_morlet(np.reshape(data.values.T, (1, data.shape[1], data.shape[0])),
                                sfreq=1./(data.index[1] - data.index[0]),
                                freqs=freqs, output=output, **kwargs)

    elif method == 'multitaper':

        from mne.time_frequency import tfr_array_multitaper
        return tfr_array_multitaper(np.reshape(data.values.T, (1, data.shape[1], data.shape[0])),
                                    sfreq=1. / (data.index[1] - data.index[0]),
                                    freqs=freqs, output=output, **kwargs)


def get_psd(data, tmin=0.0):
    # Compute spectrum
    dt = data.index[1] - data.index[0]
    n = data.shape[0]
    n_two = 1 if n == 0 else 2 ** (n - 1).bit_length()
    freqs = np.linspace(0, 1 / dt, n_two)
    spec = np.abs(np.fft.fft(data.loc[tmin:], n=n_two, axis=0))

    freqs = freqs[:int(len(spec) / 2)]
    spec = spec[:int(len(spec) / 2)]
    return freqs, spec




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""
from copy import deepcopy

from pyrates.ir.circuit import CircuitIR
from pyrates.ir.node import NodeIR
from pyrates.ir.edge import EdgeIR
from pyrates.ir.operator import OperatorIR

from pyrates.frontend._registry import register_interface

__author__ = "Daniel Rose"
__status__ = "Development"


# @register_interface
# def to_node(node_dict: dict):
#
#     order = node_dict["operator_order"]
#     variables = {}
#     # collect all information about variables and separate operator and variable keys
#     for key, var_dict in node_dict["operator_args"].items():
#         op_key, var_key = key.split("/")  # split key as 'op_key/var_key'
#         if op_key not in variables:
#             variables[op_key] = {}
#
#         variables[op_key][var_key] = var_dict
#
#     operators = []
#     for op_key in order:
#         op_dict = node_dict["operators"][op_key]
#         op_dict["name"] = op_key
#         op_dict["path"] = f"custom/{op_key}"
#         op_dict["variables"] = variables[op_key]
#         inputs = op_dict.pop("inputs")
#         for var in inputs:
#             op_dict["variables"][var]["vtype"] = "input"
#         outvar = op_dict.pop("output")
#         op_dict["variables"][outvar]["vtype"] = "output"
#         operators.append(to_operator(op_dict))
#
#     return NodeIR(operators=operators)
#
#
# @register_interface
# def to_operator(op_dict: dict):
#     from pyrates.frontend import OperatorTemplate
#     template = OperatorTemplate(**op_dict)
#
#     return template.apply()
#

@register_interface
def from_circuit(circuit: CircuitIR):
    """Reformat graph structure into a dictionary that can be saved as YAML template. The current implementation assumes
    that nodes and edges are given by as templates."""

    node_dict = {}
    for node_key, node_data in circuit.nodes(data=True):
        node = node_data["node"]
        if node.template:
            node_dict[node_key] = node.template
        else:
            # if no template is given, build and search deeper for node templates
            pass

    edge_list = []
    for source, target, edge_data in circuit.edges(data=True):
        edge_data = deepcopy(edge_data)
        edge = edge_data.pop("edge_ir")
        source = f"{source}/{edge_data['source_var']}"
        target = f"{target}/{edge_data['target_var']}"
        edge_list.append((source, target, edge.template, dict(weight=edge_data["weight"],
                                                              delay=edge_data["delay"])))

    # use Python template as base, since inheritance from YAML templates is ambiguous for circuits
    base = "CircuitTemplate"

    return dict(nodes=node_dict, edges=edge_list, base=base)




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""Visualization functionality for pyrates networks and backend simulations.
"""

# external imports
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import Union, Optional

# pyrates internal imports

# meta infos
__author__ = "Richard Gast, Daniel Rose"
__status__ = "development"


def create_cmap(name: str = None, palette_type: str = None, as_cmap: bool = True, **kwargs) -> Union[list, plt.Axes]:
    """Create a colormap or color palette object.

    Parameters
    ----------
    name
        Name of the pyrates colormap. If specified, palette_type will be ignored.
    palette_type
        Type of the seaborn color palette to use. Only necessary if no name is specified.
    as_cmap
        If true, a matplotlib colormap object will be returned. Else a seaborn color palette (list).
    kwargs
        Keyword arguments for the wrapped seaborn functions.

    Returns
    -------
    Union[list, plt.Axes]
        cmap or seaborn color palette.

    """

    from seaborn import cubehelix_palette, dark_palette, light_palette, diverging_palette, hls_palette, husl_palette, \
        color_palette, crayon_palette, xkcd_palette, mpl_palette
    import matplotlib.colors as mcolors

    if '/' in name:

        # create diverging colormap
        name1, name2 = name.split('/')
        vmin = kwargs.pop('vmin', 0.)
        vmax = kwargs.pop('vmax', 1.)
        if type(vmin) is float:
            vmin = (vmin, vmin)
        if type(vmax) is float:
            vmax = (vmax, vmax)
        kwargs1 = kwargs.pop(name1, kwargs)
        kwargs2 = kwargs.pop(name2, kwargs)
        cmap1 = create_cmap(name1, **kwargs1, as_cmap=True)
        cmap2 = create_cmap(name2, **kwargs2, as_cmap=True)
        n = kwargs.pop('n_colors', 10)
        if type(n) is int:
            n = (n, n)
        colors = np.vstack((cmap1(np.linspace(vmin[0], vmax[0], n[0])),
                            cmap2(np.linspace(vmin[1], vmax[1], n[1])[::-1])))
        return mcolors.LinearSegmentedColormap.from_list('cmap_diverging', colors)

    # extract colorrange
    if as_cmap:
        vmin = kwargs.pop('vmin', 0.)
        vmax = kwargs.pop('vmax', 1.)
        n = kwargs.pop('n_colors', 10)
        crange = np.linspace(vmin, vmax, n) if vmax-vmin < 1. else None
    else:
        crange = None

    if 'pyrates' in name:

        # create pyrates colormap
        if name == 'pyrates_red':
            cmap = cubehelix_palette(as_cmap=as_cmap, start=-2.0, rot=-0.1, **kwargs)
        elif name == 'pyrates_green':
            cmap = cubehelix_palette(as_cmap=as_cmap, start=2.5, rot=-0.1, **kwargs)
        elif name == 'pyrates_blue':
            cmap = dark_palette((210, 90, 60), as_cmap=as_cmap, input='husl', **kwargs)
        elif name == 'pyrates_yellow':
            cmap = dark_palette((70, 95, 65), as_cmap=as_cmap, input='husl', **kwargs)
        elif name == 'pyrates_purple':
            cmap = dark_palette((270, 50, 55), as_cmap=as_cmap, input='husl', **kwargs)

    else:

        # create seaborn colormap
        if palette_type == 'cubehelix':
            cmap = cubehelix_palette(as_cmap=as_cmap, **kwargs)
        elif palette_type == 'dark':
            cmap = dark_palette(as_cmap=as_cmap, **kwargs)
        elif palette_type == 'light':
            cmap = light_palette(as_cmap=as_cmap, **kwargs)
        elif palette_type == 'hls':
            cmap = hls_palette(**kwargs)
        elif palette_type == 'husl':
            cmap = husl_palette(**kwargs)
        elif palette_type == 'diverging':
            cmap = diverging_palette(as_cmap=as_cmap, **kwargs)
        elif palette_type == 'crayon':
            cmap = crayon_palette(**kwargs)
        elif palette_type == 'xkcd':
            cmap = xkcd_palette(**kwargs)
        elif palette_type == 'mpl':
            cmap = mpl_palette(name, **kwargs)
        else:
            cmap = color_palette(name, **kwargs)

    # apply colorrange
    if crange is not None:
        cmap = mcolors.LinearSegmentedColormap.from_list(name, cmap(crange))

    return cmap


def plot_timeseries(data: pd.DataFrame, variable: str = 'value', plot_style: str = 'line_plot',
                    bg_style: str = 'darkgrid', **kwargs) -> plt.Axes:
    """Plot timeseries from a data frame.

    Parameters
    ----------
    data
        Results of a pyrates simulation.
    variable
        Name of the variable to be plotted
    plot_style
        Can be either `line_plot` for plotting with seaborn.lineplot() or `ridge_plot` for using seaborn.lineplot()
        on a grid with the y-axis being separated for each population.
    bg_style
        Background style of the seaborn plot
    kwargs
        Additional key-word arguments for the seaborn function.

    Returns
    -------
    plt.Axes
        Handle of the figure axes the time-series were plotted into.

    """
    import seaborn as sb

    sb.set_style(bg_style)

    # pre-process data
    demean = kwargs.pop('demean', False)
    if demean:
        for i in range(data.shape[1]):
            data.iloc[:, i] -= np.mean(data.iloc[:, i])
            data.iloc[:, i] /= np.std(data.iloc[:, i])
    title = kwargs.pop('title', '')
    xlim, ylim = kwargs.pop('xlim', None), kwargs.pop('ylim', None)

    # Convert the dataframe to long-form or "tidy" format if necessary
    if type(data) is pd.Series:
        data_tmp = pd.DataFrame(data=data.values, columns=[variable], index=data.index)
    else:
        data_tmp = data.copy()
    idx = kwargs.pop('tmin', data_tmp.index[0])
    data_tmp = data_tmp.loc[idx:, :]
    data_tmp['time'] = data_tmp.index
    df = pd.melt(data_tmp,
                 id_vars='time',
                 var_name='node',
                 value_name=variable)

    # create color palette
    if 'cmap' in kwargs:
        cmap = kwargs.pop('cmap')
    else:
        col_pal_args = ['start', 'rot', 'gamma', 'hue', 'light', 'dark', 'reverse', 'n_colors']
        col_pal_defs = [0., 0.4, 1.0, 0.8, 0.85, 0.15, True, data_tmp.shape[1] - 1]
        kwargs_tmp = {}
        for arg, default in zip(col_pal_args, col_pal_defs):
            kwargs_tmp[arg] = kwargs.pop(arg, default)
        cmap = sb.cubehelix_palette(**kwargs_tmp)

    if 'ax' not in kwargs.keys():
        _, ax = plt.subplots()
        kwargs['ax'] = ax

    if plot_style == 'line_plot':

        # simple timeseries plot
        if 'ci' not in kwargs:
            kwargs['ci'] = None
        ylabel = kwargs.pop('ylabel', df.columns.values[0] if len(df.columns.values) == 1 else df.columns.values[0][-1])
        xlabel = kwargs.pop('xlabel', 'time')
        ax = sb.lineplot(data=df, x='time', y=variable, hue='node', palette=cmap, **kwargs)
        ax.set_title(title)
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        if xlim:
            ax.set_xlim(xlim)
        if ylim:
            ax.set_ylim(ylim)

    elif plot_style == 'ridge_plot':

        # create facet grid
        grid_args = ['col_wrap', 'sharex', 'sharey', 'height', 'aspect', 'row_order', 'col_order',
                     'dropna', 'legend_out', 'margin_titles', 'xlim', 'ylim', 'gridspec_kws', 'size',
                     'subplot_kws']
        kwargs_tmp = {}
        for key in kwargs.copy().keys():
            if key in grid_args:
                kwargs_tmp[key] = kwargs.pop(key)
        facet_hue = kwargs.pop('facet_hue', 'node')
        facet_row = kwargs.pop('facet_row', 'node')
        ax = sb.FacetGrid(df, row=facet_row, hue=facet_hue, palette=cmap, **kwargs_tmp)
        plt.close(plt.figure(plt.get_fignums()[-2]))

        # map line plots
        ax.map(sb.lineplot, 'time', variable, ci=None)
        ax.map(plt.axhline, y=0, lw=2, clip_on=False)

        # labeling args
        label_args = ['fontsize']
        kwargs_tmp = {}
        for key in kwargs.copy().keys():
            if key in label_args:
                kwargs_tmp[key] = kwargs.pop(key)

        # Define and use a simple function to label the plot in axes coordinates
        def label(x, color, label):
            ax_tmp = plt.gca()
            ax_tmp.text(0, .1, label, fontweight="bold", color=color,
                        ha="left", va="center", transform=ax_tmp.transAxes, **kwargs_tmp)

        ax.map(label, 'time')

        # Set the subplots to overlap
        hspace = kwargs.pop('hspace', -.05)
        ax.fig.subplots_adjust(hspace=hspace)

        # Remove axes details that don't play well with overlap
        ax.set_titles("")
        ax.set(yticks=[])
        ax.despine(bottom=True, left=True)

    else:

        raise ValueError(f'Plot style is not supported by this function: {plot_style}. Check the documentation of the '
                         f'argument `plot_style` for valid options.')

    return ax


def plot_trajectory(data: pd.DataFrame, x: str, y: str, z: Optional[str] = None, ax: plt.Axes = None, **kwargs):
    """Creates 2D or 3D plot of trajectory of state variables x, y and possibly z.

    Parameters
    ----------
    data
    x
    y
    z
    ax
    kwargs

    Returns
    -------

    """
    if z:
        from mpl_toolkits.mplot3d import Axes3D
        if not ax:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
        ax.plot(xs=data[x], ys=data[y], zs=data[z], **kwargs)
        return ax
    else:
        import seaborn as sb
        return sb.lineplot(x=x, y=y, data=data, **kwargs)


def plot_connectivity(fc: Union[np.ndarray, pd.DataFrame], threshold: Optional[float] = None, plot_style: str = 'heatmap',
                      bg_style: str = 'whitegrid', node_order: Optional[list] = None, auto_cluster: bool = False,
                      **kwargs) -> plt.Axes:
    """Plot functional connectivity between nodes in backend.

    Parameters
    ----------
    fc
        Pandas dataframe containing or numpy array containing the functional connectivities.
    threshold
        Connectivtiy threshold to be applied (only connectivities larger than the threshold will be shown).
    plot_style
        Can either be `heatmap` for plotting with seaborn.heatmap or `circular_graph` for plotting with
         mne.viz.plot_connectivity_circle. Check out the respective function docstrings for information on
         their arguments (can be passed to kwargs).
    bg_style
        Only relevant if plot_style == heatmap. Then this will define the style of the background of the plot.
    node_order
        Order in which the nodes should appear in the plot.
    auto_cluster
        If true, automatic cluster detection will be used to arange the nodes
    kwargs
        Additional arguments for the fc calculation or fc plotting that can be passed.

    Returns
    -------
    plt.Axes
        Handle of the axis the plot was created in.

    """

    import seaborn as sb

    # turn fc into dataframe if necessary
    if type(fc) is np.ndarray:
        rows = kwargs.pop('yticklabels') if 'yticklabels' in kwargs.keys() else [str(i) for i in range(fc.shape[0])]
        cols = kwargs.pop('xticklabels') if 'xticklabels' in kwargs.keys() else [str(i) for i in range(fc.shape[0])]
        fc = pd.DataFrame(fc, index=[str(r) for r in rows], columns=[str(c) for c in cols])

    # apply threshold
    if threshold:
        fc[fc < threshold] = 0.

    # cluster the columns
    #####################

    if auto_cluster:

        idx_r = [i for i in range(fc.shape[0])]
        idx_c = [i for i in range(fc.shape[1])]

        # Create a categorical color palette for node groups
        col_pal_args = ['h', 's', 'l']
        kwargs_tmp = {}
        for key in kwargs.keys():
            if key in col_pal_args:
                kwargs_tmp[key] = kwargs.pop(key)
        node_pal = sb.husl_palette(len(idx_c), **kwargs_tmp)
        nodes = fc.columns.values
        node_lut = dict(zip(map(str, nodes), node_pal))

        # Convert the palette to vectors that will be drawn on the side of the fc plot
        node_colors = pd.Series(nodes, index=fc.columns).map(node_lut)

    elif node_order:

        idx_c = [node_order.index(n) for n in fc.columns.values]
        idx_r = [i for i in range(fc.shape[0])]

    else:

        idx_r = [i for i in range(fc.shape[0])]
        idx_c = [i for i in range(fc.shape[1])]

    fc = fc.iloc[idx_r, idx_c]

    # plot the functional connectivities
    ####################################

    # choose plot style
    if plot_style == 'heatmap':

        # seaborn plot
        if 'xticklabels' not in kwargs:
            kwargs['xticklabels'] = fc.columns.values[idx_c]
        if 'yticklabels' not in kwargs:
            kwargs['yticklabels'] = fc.index[idx_r]

        sb.set_style(bg_style)

        if auto_cluster:
            ax = sb.clustermap(data=fc, row_colors=node_colors, col_colors=node_colors, **kwargs)
        else:
            ax = sb.heatmap(fc, **kwargs)
            # ax.invert_yaxis()

    elif plot_style == 'circular_graph':

        # mne python plot
        from mne.viz import circular_layout, plot_connectivity_circle

        # get node order for node layout
        node_names = fc.columns.values
        if auto_cluster:
            cluster_args = ['method', 'metric', 'z_score', 'standard_scale']
            kwargs_tmp = {}
            for key in kwargs.keys():
                if key in cluster_args:
                    kwargs_tmp[key] = kwargs.pop(key)
            clust_map = sb.clustermap(data=fc, row_colors=node_colors, col_colors=node_colors, **kwargs_tmp)
            node_order = [node_names[idx] for idx in clust_map.dendrogram_row.reordered_ind]
        elif not node_order:
            node_order = list(node_names)

        # create circular node layout
        kwargs_tmp = {}
        layout_args = ['start_pos', 'start_between', 'group_boundaries', 'group_sep']
        for key in kwargs.keys():
            if key in layout_args:
                kwargs_tmp[key] = kwargs.pop(key)
        node_angles = circular_layout(node_names, node_order, **kwargs_tmp)

        # plot the circular graph
        ax = plot_connectivity_circle(fc.values, node_names, node_angles=node_angles, **kwargs)

    else:

        raise ValueError(f'Plot style is not supported by this function: {plot_style}. Check the documentation of the '
                         f'argument `plot_style` for valid options.')

    return ax


def plot_phase(data: pd.DataFrame, bg_style: str = 'whitegrid', **kwargs):
    """Plot phase of populations in a polar plot.

    Parameters
    ----------
    data
        Long (tidy) format dataframe containing fields `node`, `phase` and `amplitude`.
    bg_style
        Background style of the plot.
    kwargs
        Additional keyword args to be passed to `seaborn.FacetGrid` or `seaborn.scatterplot`

    Returns
    -------
    sb.FacetGrid
        Axis handle of the created plot.

    """

    import seaborn as sb
    sb.set(style=bg_style)

    # create facet grid
    grid_args = ['col_wrap', 'sharex', 'sharey', 'height', 'aspect', 'palette', 'row_order', 'col_order',
                 'hue_order', 'hue_kws', 'dropna', 'legend_out', 'margin_titles', 'xlim', 'ylim',
                 'gridspec_kws', 'size']
    kwargs_tmp = {}
    for key in kwargs.keys():
        if key in grid_args:
            kwargs_tmp[key] = kwargs.pop(key)
    ax = sb.FacetGrid(data, hue='node', subplot_kws=dict(polar=True), despine=False, **kwargs_tmp)

    # plot phase and amplitude into polar plot
    scatter_kwargs = ['style', 'sizes', 'size_order', 'size_norm', 'markers', 'style_order', 'x_bins', 'y_bins',
                      'units', 'estimator', 'ci', 'n_boot', 'alpha', 'x_jitter', 'y_jitter', 'legend', 'ax']
    kwargs_tmp2 = {}
    for key in kwargs.keys():
        if key in scatter_kwargs:
            kwargs_tmp2[key] = kwargs.pop(key)
    ax.map(sb.scatterplot, 'phase', 'amplitude', **kwargs_tmp2)

    # plot customization
    ax_tmp = ax.facet_axis(0, 0)
    ax_tmp.set_ylim(np.min(data['amplitude']), np.max(data['amplitude']))
    ax_tmp.axes.yaxis.set_label_coords(1.15, 0.75)
    ax_tmp.set_ylabel(ax_tmp.get_ylabel(), rotation=0)
    locs, _ = plt.yticks()
    plt.yticks(locs)
    locs, labels = plt.xticks()
    labels = [np.round(l._x, 3) for l in labels]
    plt.xticks(locs, labels)

    return ax


def plot_psd(data: pd.DataFrame, fmin: float = 0., fmax: float = 100., tmin: float = 0.0, **kwargs) -> plt.Axes:
    """Plots the power-spectral density for each column in data.

    Parameters
    ----------
    data
        Dataframe with simulation results.
    fmin
        Minimum frequency to be displayed.
    fmax
        Maximum frequency to be displayed.
    tmin
        Time at which to start psd calculation.
    kwargs
        Additional keyword arguments to be passed to `mne.viz.plot_raw_psd`.

    Returns
    -------
    plt.Axes
        Handle of the created plot.

    """

    from pyrates.utility import mne_from_dataframe
    from mne.viz import plot_raw_psd

    if type(data) is pd.DataFrame and 'out_var' in data.columns.names and len(data.columns.names) > 1:

        # create heatmap
        ################

        # calculate psd's for each condition
        psd_keys = ['n_fft', 'picks', 'n_overlap', 'dB', 'estimate', 'average', 'n_jobs']
        psd_kwargs = {}
        for k in kwargs.copy():
            if k in psd_keys:
                psd_kwargs[k] = kwargs.pop(k)
        data_col = []
        col_names = []
        for col in data.columns.values:
            _ = plot_psd(data[col[:-1]], show=False, fmin=fmin, fmax=fmax, tmin=tmin, **psd_kwargs)
            pow = list(plt.gca().get_lines()[-1].get_ydata())
            freqs = list(plt.gca().get_lines()[-1].get_xdata())
            data_col.append(pow)
            col_names.append(col[:-1])
            plt.close(plt.gcf())
        df = pd.DataFrame(data=np.asarray(data_col).T, columns=col_names, index=np.round(freqs, decimals=1))

        # plot the data
        ax = plot_connectivity(df.values[::-1, :], xticklabels=df.columns.values,
                               yticklabels=df.index[::-1], plot_style='heatmap', **kwargs)
        return ax

    else:

        # create line plot
        ##################

        if len(data.shape) < 2:
            data = pd.DataFrame(data=data.values, columns=['data'], index=data.index)
        raw = mne_from_dataframe(data)
        return plot_raw_psd(raw, tmin=tmin, fmin=fmin, fmax=fmax, **kwargs).axes


def plot_tfr(data: np.ndarray, freqs: list, nodes: Optional[list] = None, separate_nodes: bool = True, **kwargs
             ) -> plt.Axes:
    """

    Parameters
    ----------
    data
        Numpy array (n x f x t) containing the instantaneous power estimates for each node (n), each frequency (f) at
        every timestep (t).
    freqs
        Frequencies of interest.
    nodes
        Nodes of interest in order of interest.
    separate_nodes
        If true, create a separate figure for each node.
    kwargs
        Additional keyword arguments to be passed to `seaborn.heatmap` or `seaborn.FacetGrid`.

    Returns
    -------
    plt.Axes
        Handle of the created plot.

    """

    import seaborn as sb

    if not nodes:
        nodes = [i for i in range(data.shape[0])]
    if 'xticklabels' not in kwargs.keys():
        if 'step_size' in kwargs.keys():
            xticks = np.round(np.arange(0, data.shape[2]) * kwargs.pop('step_size'), decimals=3)
            kwargs['xticklabels'] = [str(t) for t in xticks]
    if 'yticklabels' not in kwargs.keys():
        kwargs['yticklabels'] = [str(f) for f in freqs]

    if separate_nodes:

        # plot heatmap separately for each node
        for n in range(data.shape[0]):
            _, ax = plt.subplots()
            ax = sb.heatmap(data[n, :, :], ax=ax, **kwargs)

    else:

        # Convert the dataframe to long-form or "tidy" format
        indices = pd.MultiIndex.from_product((nodes, freqs, range(data.shape[2])), names=('nodes', 'freqs', 'time'))
        data = pd.DataFrame(data.flatten(), index=indices, columns=('values',)).reset_index()

        # create facet grid
        grid_args = ['col_wrap', 'sharex', 'sharey', 'height', 'aspect', 'palette', 'col_order',
                     'dropna', 'legend_out', 'margin_titles', 'xlim', 'ylim', 'gridspec_kws', 'size']
        kwargs_tmp = {}
        for key in kwargs.keys():
            if key in grid_args:
                kwargs_tmp[key] = kwargs.pop(key)
        ax = sb.FacetGrid(data, col='nodes', **kwargs_tmp)

        # map heatmaps to grid
        ax.map_dataframe(_draw_heatmap, 'time', 'freqs', 'values', cbar=False, square=True, **kwargs)

    return ax


def plot_network_graph(circuit, _format: str = "png", path: str = None, prog="dot", **pydot_args):
    """Simple straight plot using graphviz via pydot.

    Parameters
    ----------
    circuit
        `CircuitIR` instance to plot graph from
    _format
        output format
    path
        path to print image to. If `None` is given, will try to plot using matplotlib/IPython
    prog
        graphviz layout algorithm name. Defaults to "dot". Another recommendation is "circo". Other valid options:
        "fdp", "neato", "osage", "patchwork", "twopi", "pydot_args". See graphviz documentation for more info.

    Returns
    -------

    """

    import networkx as nx
    # import pydot

    # copy plain node names and plain edges, to avoid conflicts with graphviz
    graph = nx.MultiDiGraph()
    nodes = (node for node in circuit.graph.nodes)
    graph.add_nodes_from(nodes)
    edges = (edge for edge in circuit.graph.edges)
    graph.add_edges_from(edges)

    # pass graph to pydot
    dot_graph = nx.drawing.nx_pydot.to_pydot(graph)

    # check and format given path
    if path:
        return write_graph(path, dot_graph, _format, prog, **pydot_args)

    else:
        return show_graph(dot_graph, _format, prog, **pydot_args)


def plot_graph_with_subgraphs(circuit, _format: str = "png", path: str = None, prog="dot",
                              node_style="solid", cluster_style="rounded", **pydot_args):
    """Simple straight plot using graphviz via pydot.

    Parameters
    ----------
    circuit
        `CircuitIR` instance to plot graph from
    _format
        output format
    path
        path to print image to. If `None` is given, will try to plot using matplotlib/IPython
    prog
        graphviz layout algorithm name. Defaults to "dot". Another recommendation is "circo". Other valid options:
        "fdp", "neato", "osage", "patchwork", "twopi", "pydot_args". See graphviz documentation for more info.

    Returns
    -------

    """

    import pydot

    # copy plain node names and plain edges, to avoid conflicts with graphviz
    graph = pydot.Dot(graph_type='digraph', fontname="Verdana")
    clusters = {}
    for subcircuit in circuit.sub_circuits:
        clusters[subcircuit] = pydot.Cluster(subcircuit, label=subcircuit)

    for label, data in circuit.nodes(data=True):
        # node = data["node"]
        *subcircuit, node = label.split("/")
        node = pydot.Node(label, label=node.split(".")[0])
        node.set_style(node_style)

        if subcircuit:
            subcircuit = "/".join(subcircuit)
            cluster = clusters[subcircuit]
            cluster.add_node(node)
        else:
            graph.add_node(node)

    for cluster in clusters.values():
        cluster.set_style(cluster_style)
        graph.add_subgraph(cluster)

    for source, target, _ in circuit.edges:
        graph.add_edge(pydot.Edge(source, target))

    #
    # # pass graph to pydot
    # dot_graph = nx.drawing.nx_pydot.to_pydot(graph)
    #
    # # check and format given path
    if path:
        return write_graph(path, graph, _format, prog, **pydot_args)

    else:
        return show_graph(graph, _format, prog, **pydot_args)


def write_graph(path, dot_graph, _format, prog, **pydot_args):
    """Write DOT graph to file.

    Parameters
    ----------
    path
    dot_graph
    _format
    prog
    pydot_args

    Returns
    -------

    """
    import os
    path = os.path.normpath(path)

    # draw graph and write to file, given the format
    return dot_graph.write(path, format=_format, prog=prog, **pydot_args)


def show_graph(dot_graph, _format, prog, **pydot_args):
    """Show DOT graph using matplotlib or IPython.display

    Parameters
    ----------
    dot_graph
    _format
    prog
    pydot_args

    Returns
    -------

    """

    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg

    if _format == "png":
        # draw graph and write out postscript string
        image_bytes = dot_graph.create_png(prog=prog, **pydot_args)

        # check interactive session
        try:
            import sys
            _ = sys.ps1
        except AttributeError:
            # treat the dot output bytes as an image file
            from io import BytesIO
            bio = BytesIO()
            bio.write(image_bytes)
            bio.seek(0)
            img = mpimg.imread(bio)

            # plot the image
            imgplot = plt.imshow(img, aspect='equal')
            plt.show(block=False)
            return imgplot
        else:
            from IPython.display import Image, display

            return Image(data=image_bytes)

    else:
        raise NotImplementedError(f"No plotting option implemented for format '{_format}'")


def _draw_heatmap(*args, **kwargs):
    """Wraps seaborn.heatmap to work with long, tidy format dataframes.
    """
    import seaborn as sb

    data = kwargs.pop('data')
    d = data.pivot(index=args[1], columns=args[0], values=args[2])
    return sb.heatmap(d, **kwargs)


class Interactive2DParamPlot(object):
    def __init__(self, data_map: np.array, data_series: pd.DataFrame, x_values: np.array, y_values: np.array,
                 param_map: pd.DataFrame, tmin=0., **kwargs):
        """Creates an interactive 2D plot that allows visualization of time series using button press events

        Derive child class and change get_data() respectively to utilize this plotting method

        Parameters
        ----------
        data_map
            2D ndarray containing a value based on each column data_series, respectively.
        data_series
            DataFrame containing all data series used to crate the data map
        x_values
            ndarray containing values of data-map columns.
        y_values
            ndarray containing values of data-map rows.
        param_map
            Dataframe containing the mapping between data-series columns (index) and x/y value pairs (columns)
            (as returned by `grid_search`).
        tmin
            Starting point for time-series plots in time units (float).
        kwargs
            Additional information to access a column in data_series if necessary

        Returns
        -------

        """
        dt = kwargs.pop('dt', data_series.index[1] - data_series.index[0])
        tmin = int(tmin/dt)
        self.data = data_series.iloc[tmin:, :]
        self.x_values = x_values
        self.y_values = y_values
        self.kwargs = kwargs

        # set up param map matrix
        self.map = np.zeros(shape=data_map.shape, dtype=np.int32)
        for i in range(param_map.shape[0]):
            x, y = param_map.iloc[i, 0], param_map.iloc[i, 1]
            x_idx = np.argwhere(self.x_values == x)[0]
            y_idx = np.argwhere(self.y_values == y)[0]
            self.map[y_idx, x_idx] = i

        # Create canvas
        if 'subplots' in kwargs:
            self.fig, self.ax = kwargs.pop('subplots')
        else:
            self.fig, self.ax = plt.subplots(ncols=2, nrows=1, figsize=(12, 6), gridspec_kw={})

        # Initiate marker
        self.marker = self.ax[0].plot(0, 0, 'x', color='white', markersize='10')

        # Plot 2D data in left subplot
        num_x_ticks = kwargs.pop('num_x_ticks', 10)
        num_y_ticks = kwargs.pop('num_y_ticks', 10)
        plot_connectivity(data_map, ax=self.ax[0], yticklabels=list(np.round(y_values, decimals=2)),
                          xticklabels=list(np.round(x_values, decimals=2)), **kwargs)
        set_num_axis_ticks(ax=self.ax[0], num_x_ticks_old=data_map.shape[1], num_y_ticks_old=data_map.shape[0],
                           num_x_ticks_new=num_x_ticks, num_y_ticks_new=num_y_ticks)

        # set up grid in right subplot
        self.ax[1].grid(visible=True, color="silver")
        x, y = self.x_values[0], self.y_values[0]
        time_series = self.get_data(0, 0)
        data_min, data_max = np.min(self.data.values), np.max(self.data.values)
        data_margin = (data_max - data_min) * 0.1
        cmap = create_cmap('pyrates_purple', as_cmap=False, n_colors=1, reverse=True)
        plot_timeseries(time_series, ax=self.ax[1], ylim=[data_min-data_margin, data_max+data_margin], cmap=cmap)
        self.ax[1].set_title(f'x: {np.round(x, decimals=2)}, y: {np.round(y, decimals=2)}')

        # Call Interactive2DPlot class instance when mouse button is pressed inside the 2D plot
        self.fig.canvas.mpl_connect('button_press_event', self)

    def __call__(self, event):
        """Try to access a column in data_series using x and y values based on cursor position

        Is called on mouse button press event. Converts the current cursor coordinates inside the plot into x and y
        values based on the data in x_values and y_values. x and y values are used to access a column in data_series.
        Access of data_series can be customized in self.get_data().

        :param event:
        :return:
        """

        # Only allow button press events in the (left) 2D overview plot
        if event.inaxes != self.ax[0]:
            return

        # Reset axes
        self.marker[0].remove()

        # Transform cursor coordinates in x and y values
        x_sample = event.xdata
        y_sample = event.ydata

        # Add marker at event coordinates
        self.marker = self.ax[0].plot(x_sample, y_sample, 'x', color='white', markersize='10')

        # Update serial plot
        self.update_lineplot(int(x_sample), int(y_sample))
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def update_lineplot(self, x, y):

        line = self.ax[1].get_lines()[0]
        data = self.get_data(x, y)
        line.set_data(data.index, data.values)
        self.ax[1].set_title(f'x: {np.round(self.x_values[x],decimals=2)}, y: {np.round(self.y_values[y],decimals=2)}')
        self.ax[1].autoscale_view()

    def set_map_xlabel(self, label):
        self.ax[0].set_xlabel(label)

    def set_map_ylabel(self, label):
        self.ax[0].set_ylabel(label)

    def set_map_title(self, title):
        self.ax[0].set_title(title)

    def set_series_xlabel(self, label):
        self.ax[1].set_xlabel(label)

    def set_series_ylabel(self, label):
        self.ax[1].set_ylabel(label)

    def get_data(self, x, y, *argv, **kwargs):
        return self.data.iloc[:, self.map[y, x]]


def save_fig_as_pickle(fp, fig):
    """Save a figure object in pickle format.

    Parameters
    ----------
    fp
        /desired_dir/desired_filename.pickle
    fig
        Figure object as returned by matplotlib.pyplot.subplots

    Returns
    -------

    """
    import pickle
    pickle_out = open(fp, "wb")
    pickle.dump(fig, pickle_out)
    pickle_out.close()


def load_fig_from_pickle():
    """Opens a file dialog to select and load a *.pickle file.

    If the pickle file contains a figure object it will automatically be plotted

    Returns
    -------

    """
    import pickle
    from tkinter import Tk, filedialog

    root = Tk()
    root.withdraw()
    file_path = filedialog.askopenfilename(filetypes=(("pickle files", "*.pickle"), ("all files", "*.*")))

    try:
        pickle_in = open(file_path, "rb")
        pickle.load(pickle_in)
        plt.show()
    except (FileNotFoundError, TypeError, pickle.UnpicklingError):
        pass


def set_num_axis_ticks(ax, num_x_ticks_old, num_y_ticks_old, num_x_ticks_new=10, num_y_ticks_new=10):
    """Set the number of x and y ticks of a plot axis

    Parameters
    ----------
    ax
    num_x_ticks_old
    num_y_ticks_old
    num_x_ticks_new
    num_y_ticks_new

    Returns
    -------

    """
    step_tick_x, step_tick_y = int(num_x_ticks_old / num_x_ticks_new), int(num_y_ticks_old / num_y_ticks_new)
    for n, tick in enumerate(ax.xaxis.iter_ticks()):
        if n % step_tick_x != 0:
            tick[0].set_visible(False)
    for n, tick in enumerate(ax.yaxis.iter_ticks()):
        if n % step_tick_y != 0:
            tick[0].set_visible(False)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""This module provides parser classes and functions to parse string-based equations into symbolic representations of
operations.
"""

import math
import typing as tp
from copy import deepcopy
from numbers import Number

# external imports
from pyparsing import Literal, CaselessLiteral, Word, Combine, Optional, \
    ZeroOrMore, Forward, nums, alphas, ParserElement

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


# expression parsers (lhs/rhs of an equation)
#############################################


class ExpressionParser(ParserElement):
    """Base class for parsing mathematical expressions from a string format into a symbolic representation of the
    mathematical operation expressed by it.

    Parameters
    ----------
    expr_str
        Mathematical expression in string format.
    args
        Dictionary containing all variables and functions needed to evaluate the expression.
    backend
        Backend instance in which to parse all variables and operations.
        See `pyrates.backend.numpy_backend.NumpyBackend` for a full documentation of the backends methods and
        attributes.
    kwargs
        Additional keyword arguments to be passed to the backend functions.

    Attributes
    ----------
    lhs
        Boolean, indicates whether expression is left-hand side or right-hand side of an equation
    args
        Dictionary containing the variables of an expression
    solve
        Only relevant for lhs expressions. If True, lhs will be treated as a first-order ordinary differential equation.
    expr_str
        String representation of the mathematical expression
    expr
        Symbolic (Pyparsing-based) representation of mathematical expression.
    expr_stack
        List representation of the syntax tree of the (parsed) mathematical expression.
    expr_list
        List representation of the mathematical expression.
    op
        Operator for calculating the mathematical expression (symbolic representation).
    _op_tmp
        Helper variable for building `op`.
    ops
        Dictionary containing all mathematical operations that are allowed for a specific instance of the
        `ExpressionParser` (e.g. +, -, *, /).
    funcs
        Dictionary containing all additional functions that can be used within mathematical expressions with a specific
        instance of the `ExpressionParser` (e.g. sum(), reshape(), float32()).
    dtypes
        Dictionary containing all data-types that can be used within mathematical expressions with a specific instance
        of the `ExpressionParser` (e.g. float32, bool, int32).

    """

    def __init__(self, expr_str: str, args: dict, backend: tp.Any, **kwargs) -> None:
        """Instantiates expression parser.
        """

        # call super init
        #################

        super().__init__()

        # bind attributes to instance
        #############################

        # input arguments
        self.vars = args.copy()
        self.backend = backend
        self.parser_kwargs = kwargs

        self.lhs, self.rhs, self._diff_eq, self._assign_type, self.lhs_key = self._preprocess_expr_str(expr_str)

        # add functions from args dictionary to backend, if passed
        for key, val in args.items():
            if callable(val):
                self.backend.ops[key] = val

        # additional attributes
        self.expr_str = expr_str
        self.expr = None
        self.expr_stack = []
        self.expr_list = []
        self.op = None
        self._finished_rhs = False
        self.rhs_eval = None
        self._instantaneous = kwargs.pop('instantaneous', False)

        # define algebra
        ################

        if not self.expr:

            # general symbols
            point = Literal(".")
            comma = Literal(",")
            colon = Literal(":")
            e = CaselessLiteral("E")
            pi = CaselessLiteral("PI")

            # parentheses
            par_l = Literal("(")
            par_r = Literal(")").setParseAction(self._push_first)
            idx_l = Literal("[")
            idx_r = Literal("]")

            # basic mathematical operations
            plus = Literal("+")
            minus = Literal("-")
            mult = Literal("*")
            div = Literal("/")
            mod = Literal("%")
            dot = Literal("@")
            exp_1 = Literal("^")
            exp_2 = Combine(mult + mult)
            transp = Combine(point + Literal("T"))
            inv = Combine(point + Literal("I"))

            # numeric types
            num_float = Combine(Word("-" + nums, nums) +
                                Optional(point + Optional(Word(nums))) +
                                Optional(e + Word("-" + nums, nums)))
            num_int = Word("-" + nums, nums)

            # variables and functions
            name = Word(alphas, alphas + nums + "_$")
            func_name = Combine(name + par_l, adjacent=True)

            # math operation groups
            op_add = plus | minus
            op_mult = mult | div | dot | mod
            op_exp = exp_1 | exp_2 | inv | transp

            # logical operations
            greater = Literal(">")
            less = Literal("<")
            equal = Combine(Literal("=") + Literal("="))
            unequal = Combine(Literal("!") + Literal("="))
            greater_equal = Combine(Literal(">") + Literal("="))
            less_equal = Combine(Literal("<") + Literal("="))

            # logical operations group
            op_logical = greater_equal | less_equal | unequal | equal | less | greater

            # pre-allocations
            self.expr = Forward()
            exponential = Forward()
            index_multiples = Forward()

            # basic organization units
            index_start = idx_l.setParseAction(self._push_first)
            index_end = idx_r.setParseAction(self._push_first)
            index_comb = colon.setParseAction(self._push_first)
            arg_comb = comma.setParseAction(self._push_first)
            arg_tuple = par_l + ZeroOrMore(self.expr.suppress() + Optional(arg_comb)) + par_r
            func_arg = arg_tuple | self.expr.suppress()

            # basic computation unit
            atom = (func_name + Optional(func_arg.suppress()) + ZeroOrMore(arg_comb.suppress() + func_arg.suppress()) +
                    par_r.suppress() | name | pi | e | num_float | num_int).setParseAction(self._push_neg_or_first) | \
                   (par_l.setParseAction(self._push_last) + self.expr.suppress() + par_r).setParseAction(self._push_neg)

            # apply indexing to atoms
            indexed = atom + ZeroOrMore((index_start + index_multiples + index_end))
            index_base = (self.expr.suppress() | index_comb)
            index_full = index_base + ZeroOrMore((index_comb + index_base)) + ZeroOrMore(index_comb)
            index_multiples << index_full + ZeroOrMore((arg_comb + index_full))

            # hierarchical relationships between mathematical and logical operations
            boolean = indexed + Optional((op_logical + indexed).setParseAction(self._push_first))
            exponential << boolean + ZeroOrMore((op_exp + Optional(exponential)).setParseAction(self._push_first))
            factor = exponential + ZeroOrMore((op_mult + exponential).setParseAction(self._push_first))
            self.expr << factor + ZeroOrMore((op_add + factor).setParseAction(self._push_first))

    def parse_expr(self) -> tuple:
        """Parses string-based mathematical expression/equation.

        Returns
        -------
        tuple
            left-hand side, right-hand side and variables of the parsed equation.
        """

        # extract symbols and operations from equations right-hand side
        self.expr_list = self.expr.parseString(self.rhs)
        self._check_parsed_expr(self.rhs)

        # parse rhs into backend
        rhs = self.parse(self.expr_stack[:])

        if self.lhs_key in self.vars and not self._instantaneous and hasattr(rhs, 'short_name'):

            # create new variable for lhs update
            name = f"{self.lhs_key}_delta"
            i = 0
            name = f"{name}_{i}"
            while name in self.vars:
                i += 1
                name[-1] = i
            dtype = rhs.dtype if hasattr(rhs, 'dtype') else self.backend._float_def
            shape = rhs.shape if hasattr(rhs, 'shape') else ()

            # assign rhs to new variable
            delta = self.backend.add_var('state_var', name=name, dtype=dtype, shape=shape, **self.parser_kwargs)
            self.backend.add_op('=', delta, rhs)
            self.rhs = delta

        else:

            self.rhs = rhs

        # post rhs parsing steps
        self.rhs_eval = rhs
        self.clear()
        self._finished_rhs = True

        # extract symbols and operations from left-hand side
        self.expr_list = self.expr.parseString(self.lhs)
        self._check_parsed_expr(self.lhs)

        # parse lhs into backend
        self.lhs = self._update_lhs()

        return self.lhs, self.rhs, self.vars

    def parse(self, expr_stack: list) -> tp.Any:
        """Parse elements in expression stack into the backend.

        Parameters
        ----------
        expr_stack
            Ordered list with expression variables and operations. Needs to be processed from last to first item.

        Returns
        -------
        tp.Any
            Parsed expression stack element (object type depends on the backend).

        """

        # get next operation from stack
        op = expr_stack.pop()

        # check type of operation
        #########################

        if op == '-one':

            # multiply expression by minus one
            self.op = self.backend.add_op('*', self.parse(expr_stack), -1, **self.parser_kwargs)

        elif op in ["*=", "/=", "+=", "-=", "="]:

            # collect rhs
            op1 = self.parse(expr_stack)

            # collect lhs
            indexed_lhs = True if "]" in expr_stack else False
            op2 = self.parse(expr_stack)

            # combine elements via mathematical/boolean operator
            if indexed_lhs:
                self.op = self._apply_idx(op=op2[0], idx=op2[1], update=op1, update_type=op, **self.parser_kwargs)
            else:
                self.op = self.backend.add_op(op, op2, op1, **self.parser_kwargs)

        elif op in "+-/**^@<=>=!==%":

            # collect elements to combine
            op2 = self.parse(expr_stack)
            op1 = self.parse(expr_stack)

            # combine elements via mathematical/boolean operator
            self.op = self.backend.add_op(op, op1, op2, **self.parser_kwargs)

        elif ".T" == op or ".I" == op:

            # transpose/invert expression
            self.op = self.backend.add_op(op, self.parse(expr_stack), **self.parser_kwargs)

        elif op == "]":

            # parse indices
            indices = []
            while len(expr_stack) > 0 and expr_stack[-1] != "[":
                index = []
                while len(expr_stack) > 0 and expr_stack[-1] not in ",[":
                    if expr_stack[-1] == ":":
                        index.append(expr_stack.pop())
                    else:
                        try:
                            int(expr_stack[-1])
                            index.append(expr_stack.pop())
                        except ValueError:
                            tmp = self._finished_rhs
                            self._finished_rhs = False
                            index.append(self.parse(expr_stack))
                            self._finished_rhs = tmp
                indices.append(index[::-1])
                if expr_stack[-1] == ",":
                    expr_stack.pop()
            expr_stack.pop()

            # build string-based representation of idx
            if 'idx' not in self.vars.keys():
                self.vars['idx'] = {}
            idx = ""
            i = 0
            for index in indices[::-1]:
                for ind in index:
                    if type(ind) == str:
                        idx += ind
                    elif isinstance(ind, Number):
                        idx += f"{ind}"
                    else:
                        self.vars['idx'][f'idx_var_{i}'] = ind
                        idx += f"idx_var_{i}"
                    i += 1
                idx += ","
            idx = idx[0:-1]

            # extract variable and apply idx if its a rhs variable. Else return variable and index
            if self._finished_rhs:
                op = expr_stack.pop(-1)
                if op in self.vars:
                    op_to_idx = self.vars[op]
                else:
                    op_to_idx = self.parse([op])
                self.op = (op_to_idx, idx)
            else:
                op_to_idx = self.parse(expr_stack)
                op_idx = self._apply_idx(op_to_idx, idx, **self.parser_kwargs)
                self.op = op_idx

        elif op == "PI":

            # return float representation of pi
            self.op = math.pi

        elif op == "E":

            # return float representation of e
            self.op = math.e

        elif op in self.vars:

            # extract constant/variable from args dict
            self.op = self.vars[op]

        elif any(["float" in op, "bool" in op, "int" in op, "complex" in op]):

            expr_stack.pop(-1)

            # extract data type
            try:
                self.op = self.backend.add_op('cast', self.parse(expr_stack), op[0:-1], **self.parser_kwargs)
            except AttributeError:
                raise AttributeError(f"Datatype casting error in expression: {self.expr_str}. "
                                     f"{op[0:-1]} is not a valid data-type for this parser.")

        elif op[-1] == "(":

            expr_stack.pop(-1)

            # parse arguments
            args = []
            while len(expr_stack) > 0:
                args.append(self.parse(expr_stack))
                if len(expr_stack) == 0 or expr_stack[-1] != ",":
                    break
                else:
                    expr_stack.pop()

            # apply function to arguments
            try:
                if len(args) == 1:
                    self.op = self.backend.add_op(op[0:-1], args[0], **self.parser_kwargs)
                else:
                    self.op = self.backend.add_op(op[0:-1], *tuple(args[::-1]), **self.parser_kwargs)
            except KeyError:
                raise KeyError(
                    f"Undefined function in expression: {self.expr_str}. {op[0:-1]} needs to be provided "
                    f"in arguments dictionary.")

        elif op == ")":

            # check whether expression in parenthesis is a group of arguments to a function
            start_par = -1
            found_end = 0
            while found_end < 1:
                if "(" in expr_stack[start_par]:
                    found_end += 1
                if ")" in expr_stack[start_par]:
                    found_end -= 1
                start_par -= 1
            if "," in expr_stack[start_par+1:]:

                args = []
                while True:
                    args.append(self.parse(expr_stack))
                    if expr_stack[-1] == ",":
                        expr_stack.pop(-1)
                    elif expr_stack[-1] == "(":
                        expr_stack.pop(-1)
                        break
                    else:
                        break
                self.op = args[::-1]

            else:

                self.op = self.parse(expr_stack)
                expr_stack.pop(-1)

        elif any([op == "True", op == "true", op == "False", op == "false"]):

            # return boolean
            self.op = True if op in "Truetrue" else False

        elif "." in op:

            self.op = float(op)

        elif op.isnumeric():

            self.op = int(op)

        elif op[0].isalpha():

            if self._finished_rhs:

                if op == 'rhs':
                    self.op = self.rhs
                else:
                    shape = self.rhs_eval.shape if hasattr(self.rhs_eval, 'shape') else ()
                    dtype = self.rhs_eval.dtype if hasattr(self.rhs_eval, 'dtype') else type(self.rhs_eval)
                    self.op = self.backend.add_var(vtype='state_var', name=op, shape=shape, dtype=dtype)
                self.vars[op] = self.op

            else:

                raise ValueError(f"Undefined variable detected in expression: {self.expr_str}. {op} was not found "
                                 f"in the respective arguments dictionary.")

        else:

            raise ValueError(f"Undefined operation detected in expression: {self.expr_str}. {op} cannot be "
                             f"interpreted by this parser.")

        return self.op

    def clear(self):
        """Clears expression list and stack.
        """
        self.expr_list.clear()
        self.expr_stack.clear()

    def _update_lhs(self):
        """Applies update to left-hand side of equation. For differential equations, different solving schemes are
        available.
        """

        # extract update parameters
        solver = self.parser_kwargs.pop('solver', 'euler')
        diff_eq = self._diff_eq

        # update left-hand side of equation
        ###################################

        if diff_eq:

            if solver == 'euler':

                # use explicit forward euler
                self.backend.next_layer()
                op = self.parse(self.expr_stack + ['dt', 'rhs', '*', '+='])
                self.backend.previous_layer()

            else:

                raise ValueError(f'Wrong solver type: {solver}. '
                                 f'Please check the docstring of this function for available solvers.')

        elif self._instantaneous:

            # simple instantaneous update
            op = self.parse(self.expr_stack + ['rhs', self._assign_type])

        else:

            # simple non-instantaneous update
            self.backend.next_layer()
            op = self.parse(self.expr_stack + ['rhs', self._assign_type])
            self.backend.previous_layer()

        return op

    def _preprocess_expr_str(self, expr: str) -> tuple:
        """Turns differential equations into simple algebraic equations using a certain solver scheme and extracts
        left-hand side, right-hand side and update type of the equation.

        Parameters
        ----------
        expr
            Equation in string format.

        Returns
        -------
        tuple
            Contains left hand side, right hand side and left hand side update type
        """

        # collect equation specifics
        ############################

        # split equation into lhs and rhs and assign type
        lhs, rhs, assign_type = split_equation(expr)

        if not assign_type:
            return self._preprocess_expr_str(f"x = {expr}")

        # for the left-hand side, check whether it includes a differential operator
        if "d/dt" in lhs:
            diff_eq = True
            lhs_split = lhs.split('*')
            lhs = "".join(lhs_split[1:])
        elif "'" in lhs:
            diff_eq = True
            lhs = lhs.replace("'", "")
        elif "d" in lhs and "/dt" in lhs:
            diff_eq = True
            lhs = lhs.split('/dt')[0]
            lhs = lhs.replace("d", "", count=1)
        else:
            diff_eq = False

        # get clean name of lhs
        lhs_key = lhs.split('[')[0]
        lhs_key = lhs_key.replace(' ', '')
        lhs = lhs.replace(' ', '')

        # store equation specifics
        if diff_eq and assign_type != '=':
            raise ValueError(f'Wrong assignment method for equation: {expr}. '
                             f'A differential equation cannot be combined with an assign type other than `=`.')

        return lhs, rhs, diff_eq, assign_type, lhs_key

    def _push_first(self, strg, loc, toks):
        """Push tokens in first-to-last order to expression stack.
        """
        self.expr_stack.append(toks[0])

    def _push_neg(self, strg, loc, toks):
        """Push negative one multiplier if on first position in toks.
        """
        if toks and toks[0] == '-':
            self.expr_stack.append('-one')

    def _push_neg_or_first(self, strg, loc, toks):
        """Push neg one multipler to expression stack if on first position in toks, else push toks from first-to-last.
        """
        if toks and toks[0] == '-':
            self.expr_stack.append('-one')
        else:
            self.expr_stack.append(toks[0])

    def _push_last(self, strg, loc, toks):
        """Push tokens in last-to-first order to expression stack.
        """
        self.expr_stack.append(toks[-1])

    def _apply_idx(self, op: tp.Any, idx: tp.Any, update: tp.Optional[tp.Any] = None,
                   update_type: tp.Optional[str] = None, **kwargs) -> tp.Any:
        """Apply index idx to operation op.

        Parameters
        ----------
        op
            Operation to be indexed.
        idx
            Index to op.
        update
            Update to apply to op at idx.
        update_type
            Type of left-hand side update (e.g. `=` or `+=`).
        kwargs
            Additional keyword arguments to be passed to the indexing functions.

        Returns
        -------
        tp.Any
            Result of applying idx to op.

        """

        kwargs.update(self.parser_kwargs)

        # get constants/variables that are part of the index
        args = []
        if idx in self.vars['idx']:
            idx = self.vars['idx'].pop(idx)
        if type(idx) is str:
            idx_old = idx
            idx = []
            for idx_tmp in idx_old.split(','):
                for idx_tmp2 in idx_tmp.split(':'):
                    idx.append(idx_tmp2)
                    if idx_tmp2 in self.vars['idx']:
                        idx_var = self.vars['idx'].pop(idx_tmp2)
                        if not hasattr(idx_var, 'short_name'):
                            if hasattr(idx_var, 'shape') and tuple(idx_var.shape):
                                idx_var = idx_var[0]
                            idx[-1] = f"{idx_var}"
                        else:
                            idx[-1] = idx_var.short_name
                            args.append(idx_var)
                    idx.append(':')
                idx.pop(-1)
                idx.append(',')
            idx.pop(-1)
            idx = "".join(idx)

        return self.backend.apply_idx(op, idx, update, update_type, *tuple(args))

    def _check_parsed_expr(self, expr_str: str) -> None:
        """check whether parsing of expression string was successful.

        Parameters
        ----------
        expr_str
            Expression that has been attempted to be parsed.
        """

        for sub_str in sorted(self.expr_stack, key=len)[::-1]:
            if sub_str == 'E':
                sub_str = 'e'
            expr_str = expr_str.replace(sub_str, "")
        expr_str = expr_str.replace(" ", "")
        expr_str = expr_str.replace("(", "")
        expr_str = expr_str.replace(")", "")
        expr_str = expr_str.replace("-", "")
        if len(expr_str) > 0:
            raise ValueError(f"Error while parsing expression: {self.expr_str}. {expr_str} could not be parsed.")

    @staticmethod
    def _compare(x: tp.Any, y: tp.Any) -> bool:
        """Checks whether x and y are equal or not.
        """
        test = x == y
        if hasattr(test, 'shape'):
            test = test.any()
        return test


def parse_equation_system(equations: list, equation_args: dict, backend: tp.Any, solver='euler', **kwargs) -> dict:
    """Parses a system (list) of equations into the backend. Transforms differential equations into the appropriate set
    of algebraic equations (depends on the chosen solver).

    Parameters
    ----------
    equations
        Collection of equations that describe the dynamics of the nodes and edges.
    equation_args
        Key-value pairs of arguments needed for parsing the equations.
    backend
        Backend instance to parse the equations into.
    solver
        Type of solving algorithm to be applied to the differential equations in the system. Available solvers are:
        - `euler` for the explicit euler method
        - `midpoint` for the midpoint method
        - `rk23` for the Runge-Kutta algorithm of 2nd (3rd) order
    kwargs
        Additional keyword arguments to be passed to the backend methods.

    Returns
    -------
    dict
        The updated equations args (in-place manipulation of all variables in equation_args happens during
        equation parsing).

    """

    update_num = 0

    if solver == 'midpoint':

        print("EXPERIMENTAL FEATURE WARNING: Using the midpoint method for solving differential equations is not fully "
              "tested yet and should be used with caution.")

        # first rhs evaluation
        update_num += 1
        state_vars = {key: var for key, var in equation_args.items()
                      if (type(var) is dict) and ('vtype' in var) and (var['vtype'] == 'state_var')}
        state_vars_orig = dict(state_vars.items())
        equations_tmp, state_vars = update_lhs(deepcopy(equations), state_vars, update_num, state_vars_orig)
        equation_args.update(state_vars)
        updates, equation_args = parse_equations(equations=equations_tmp, equation_args=equation_args, backend=backend,
                                                 **kwargs)
        backend.add_layer()

        # second rhs evaluation + combination of the two
        updates.update({key: arg for key, arg in equation_args.items() if 'inputs' in key})
        equations, _ = update_rhs(equations, updates, update_num, "(var_placeholder + 0.5*update_placeholder)")
        _, equation_args = parse_equations(equations=equations, equation_args=equation_args, backend=backend, **kwargs)

    elif solver == 'rk23':

        print("EXPERIMENTAL FEATURE WARNING: Using the Runge-Kutta 23 algorithm for solving differential equations is "
              "not fully tested yet and should be used with caution.")

        # first rhs evaluation
        update_num += 1
        state_vars = {key: var for key, var in equation_args.items()
                      if (type(var) is dict) and ('vtype' in var) and (var['vtype'] == 'state_var')}
        state_vars_orig = dict(state_vars.items())
        equations_tmp, state_vars = update_lhs(deepcopy(equations), state_vars, update_num, state_vars_orig)
        for var_orig in state_vars_orig.copy().keys():
            if not any([var_orig in var for var in state_vars]):
                state_vars_orig.pop(var_orig)
        equation_args.update(state_vars)
        updates, equation_args = parse_equations(equations=equations_tmp, equation_args=equation_args, backend=backend,
                                                 **kwargs)
        backend.add_layer()

        # second rhs evaluation
        updates.update({key: arg for key, arg in equation_args.items() if 'inputs' in key})
        equations_tmp, updates = update_rhs(deepcopy(equations), updates, update_num,
                                                                       "(var_placeholder + 0.5*update_placeholder)")
        #equation_args = update_equation_args(equation_args, updates)
        state_vars = {key: var for key, var in equation_args.items()
                      if any([orig_key in key for orig_key in state_vars_orig])}
        update_num += 1
        equations_tmp, state_vars = update_lhs(equations_tmp, state_vars, update_num, state_vars_orig)
        equation_args.update(state_vars)
        updates_new, equation_args = parse_equations(equations=equations_tmp, equation_args=equation_args,
                                                     backend=backend, **kwargs)
        backend.add_layer()

        # third rhs evaluation
        updates.update({key: arg for key, arg in equation_args.items() if 'inputs' in key})
        updates.update(updates_new)
        equations, updates = update_rhs(equations, updates, update_num,
                                        "(var_placeholder - var_placeholder_1 + 2*update_placeholder)")
        #equation_args = update_equation_args(equation_args, updates)
        state_vars = {key: var for key, var in equation_args.items()
                      if any([orig_key in key for orig_key in state_vars_orig])}
        update_num += 1
        equations, state_vars = update_lhs(equations, state_vars, update_num, state_vars_orig)
        equation_args.update(state_vars)
        updates, equation_args = parse_equations(equations=equations, equation_args=equation_args, backend=backend,
                                                 **kwargs)
        backend.add_layer()

        # combination of 3 rhs evaluations a'la rk23
        #equation_args = update_equation_args(equation_args, updates)
        equations = []
        for var_info in state_vars_orig:
            node, op, var = var_info.split('/')
            equations.append((f'{var} += ({var}_upd_1 + 4.*{var}_upd_2 + {var}_upd_3)/6.', f"{node}/{op}"))
        _, equation_args = parse_equations(equations=[equations], equation_args=equation_args, backend=backend,
                                           edge_layer=False, **kwargs)

    elif solver == 'euler':

        _, equation_args = parse_equations(equations=equations, equation_args=equation_args, backend=backend, **kwargs)

    else:

        valid_solvers = ['euler', 'midpoint', 'rk23']
        raise ValueError(f'Invalid solver: {solver}. Valid differential equation solvers are: {valid_solvers}.')

    return equation_args


def parse_equations(equations: list, equation_args: dict, backend: tp.Any, **kwargs
                    ) -> tuple:
    """Parses list of equations into the backend.

    Parameters
    ----------
    equations
        List of equations to parse into the backend.
    equation_args
        Dictionary with arguments needed for the equations.
    backend
        Backend instance that the equations should be parsed into.

    Returns
    -------
    tuple
        Two dictionaries. The first one contains all left-hand side variables for which update operations were created
        according to their right-hand sides. The second one is the updated `equation_args` dictionary that was
        passed to this function.
    """
    updates = {}

    for layer in equations:

        coupled = is_coupled(layer)

        for eq, scope in layer:

            # parse arguments
            #################

            # extract operator variables from equation args
            op_args = {key.split('/')[-1]: var for key, var in equation_args.items() if scope in key}
            inputs = op_args['inputs'] if 'inputs' in op_args else {}
            unprocessed_inputs = []
            for key, inp in inputs.items():
                if inp not in equation_args:
                    raise KeyError(inp)
                inp_tmp = equation_args[inp]
                # TODO: For some reason, some equation args are added after this point that I do not understand.
                op_args[key] = inp_tmp
                if type(inp_tmp) is dict and 'vtype' in inp_tmp:
                    unprocessed_inputs.append(key)

            # parse operator variables in backend
            args_tmp = {}
            for key, arg in op_args.items():
                if type(arg) is dict and 'vtype' in arg:
                    args_tmp[key] = arg
            args_tmp = parse_dict(args_tmp, backend, scope=scope, **kwargs)
            op_args.update(args_tmp)
            op_args['dt'] = equation_args['all/all/dt']

            # parse equation
            ################

            diff_eq = is_diff_eq(eq)

            if not diff_eq and not coupled:

                # apply update to lhs variable of equation instantaneously
                parser = ExpressionParser(expr_str=eq, args=op_args, backend=backend, scope=scope, instantaneous=True,
                                          **kwargs.copy())
                parser.parse_expr()

            else:

                # apply save, non-instantaneous update to lhs variable
                parser = ExpressionParser(expr_str=eq, args=op_args, backend=backend, scope=scope, instantaneous=False,
                                          **kwargs.copy())
                parser.parse_expr()
                if diff_eq:
                    updates[f"{scope}/{parser.lhs_key}"] = parser.vars[parser.lhs_key]

            # udpate equations args
            #######################

            # save backend variables to equation args
            for key, var in parser.vars.items():
                if key != "inputs" and key != "rhs" and key != "dt":
                    #  TODO: this step adds newly created variables to the equation args like the output variable to a
                    #   collection variable
                    equation_args[f"{scope}/{key}"] = var

            # save previously unprocessed input variables to equation args
            for key, inp in inputs.items():
                if key in unprocessed_inputs:
                    equation_args[inp] = parser.vars[key]

        # go to next layer in backend
        backend.add_layer()

    return updates, equation_args


def update_rhs(equations: list, equation_args: dict, update_num: int, update_str: str) -> tuple:
    """Update the right-hand side of all equations according to `update_str` and `update_num`. All state-variable
    occurrences will be replaced with the expression in the `update_str` template. Convenience function for differential
    equation solver that involve multiple partial updates of the state variables.

    Parameters
    ----------
    equations
        List of equations to be updated.
    equation_args
        Key-argument pairs of all relevant variables which occur in the equations.
    update_num
        Number of the partial update step for which the equations should be updated.
    update_str
        Template for the state variable replacement procedure. Should contain the following character strings:
        - `var_placeholder` will be replaced with the name of the state variables.
        - `var_placeholder_i` for partial updates of the state variables with `i` being a counter that needs to be
           replaced with the appropriate number of the partial update. Should be included for each partial update from
           i=1 to i=`update_num`.
        - `update_placeholder` for the position of the new, updated variable.

    Returns
    -------
    tuple
        List of the updated equations and dictionary with the equation arguments.

    """

    # collect variable updates from earlier rhs evaluations
    #######################################################

    var_updates = {}
    if update_num > 1:

        for key, arg in equation_args.items():

            # extraction of variable name
            node, op, var = key.split('/')
            if f"_upd_{update_num}" in var:
                var = var.replace(f"_upd_{update_num}", "")

            if "_upd_" in var:

                # find variable update number and cut of variable update identifier from variable name
                idx = int(var[-1])
                var_tmp = var[:-6]

                # indicate which variable placeholder to replace with this variable update below
                if var_tmp in var_updates:
                    var_updates[var_tmp].append((f'var_placeholder_{idx}', var))
                else:
                    var_updates[var_tmp] = [(f'var_placeholder_{idx}', var)]

    # integrate previous rhs evaluations into rhs equations
    #######################################################

    updated_args = {}
    while equation_args:

        key, arg = equation_args.popitem()

        if 'inputs' in key:

            # process input variables
            for var, arg_tmp in arg.copy().items():

                # extract variable name
                if f"_upd_{update_num}" in var:
                    var = var.replace(f"_upd_{update_num}", "")

                if "_upd_" not in var:

                    # create new variable name with update identifier
                    new_var = f"{var}_upd_{update_num}"
                    arg_tmp = arg_tmp.split('/')
                    arg_tmp[-1] = f"{arg_tmp[-1]}_upd_{update_num}"
                    arg_tmp = "/".join(arg_tmp)

                    if arg_tmp in equation_args or arg_tmp in updated_args:

                        # add variable update to inputs field
                        arg[new_var] = arg_tmp

                        # individualize the replacement template with variable name infos
                        replace_str = update_str.replace('update_placeholder', new_var)
                        if var in var_updates:
                            for placeholder, var_tmp in var_updates[var]:
                                replace_str = replace_str.replace(placeholder, var_tmp)
                        else:
                            for i in range(1, update_num):
                                replace_str = replace_str.replace(f'var_placeholder_{i}', f'{var}_upd_{i}')
                        replace_str = replace_str.replace('var_placeholder', var)

                        # go through equations and replace variable names with the individualized replacement template
                        for i, layer in enumerate(equations.copy()):
                            for j, (eq, scope) in enumerate(layer):
                                lhs, rhs, assign = split_equation(eq)
                                if replace_str not in rhs:
                                    rhs = replace(rhs, var, replace_str)
                                equations[i][j] = (f"{lhs} {assign} {rhs}", scope)

        else:

            # extract variable name
            node, op, var = key.split('/')
            if f"_upd_{update_num}" in var:
                var = var.replace(f"_upd_{update_num}", "")

            if "_upd_" in var:

                # save the variable to the arguments dictionary
                updated_args[f"{node}/{op}/{var}"] = arg

            else:

                # create new variable name and save the variable to the arguments dictionary
                new_var = f"{var}_upd_{update_num}"
                updated_args[f"{node}/{op}/{new_var}"] = arg

                # individualize the replacement template
                replace_str = update_str.replace('update_placeholder', new_var)
                if var in var_updates:
                    for placeholder, var_tmp in var_updates[var]:
                        replace_str = replace_str.replace(placeholder, var_tmp)
                replace_str = replace_str.replace('var_placeholder', var)

                # go through equations and replace variable occurences with the individualized replacement template
                for i, layer in enumerate(equations.copy()):
                    for j, (eq, scope) in enumerate(layer):
                        lhs, rhs, assign = split_equation(eq)
                        if replace_str not in rhs:
                            rhs = replace(rhs, var, replace_str)
                        equations[i][j] = (f"{lhs} {assign} {rhs}", scope)

    return equations, updated_args


def update_lhs(equations: list, equation_args: dict, update_num: int, var_dict: dict) -> tuple:
    """Update the left-hand side of all equations according to `update_num`. An update identifier will be added to all
    left-hand side state-variable occurences. Convenience function for differential equation solver that involve
    multiple partial updates of the state variables.

    Parameters
    ----------
    equations
        Equations, whose left-hand sides should be updated.
    equation_args
        Key-argument pairs including all relevant left-hand side variables.
    update_num
        Number of the partial udpate of the left-hand side variables.
    var_dict
        Key-argument pairs including the configurations of all state variables (like shape, dtype, vtype and value).


    Returns
    -------
    tuple
        List of the updated equations and dictionary with the equation arguments.
    """

    updated_args = {}
    while equation_args:

        # extract variable
        key, arg = equation_args.popitem()
        node, op, var = key.split('/')

        if "_upd_" in var and f"_upd_{update_num-1}" not in var:

            # add previous updates to the arguments dictionary
            updated_args[key] = arg

        else:

            # create new variable name
            var = var.replace(f"_upd_{update_num-1}", "")
            new_var = f"{var}_upd_{update_num}"

            # go through equations and replace the left-hand side variables of differential equations
            add_to_args = False
            for i, layer in enumerate(equations.copy()):
                for j, (eq, scope) in enumerate(layer):
                    if node in scope and op in scope:
                        lhs, rhs, _ = split_equation(eq)
                        if var in lhs and new_var in replace(lhs, var, new_var):
                            de = False
                            if "d/dt" in lhs:
                                de = True
                                add_to_args = True
                                lhs = lhs.replace("d/dt", "")
                                lhs = lhs.replace("*", "")
                                lhs = lhs.replace(" ", "")
                            elif "'" in lhs:
                                de = True
                                add_to_args = True
                                lhs = lhs.replace("'", "")
                                lhs = lhs.replace(" ", "")
                            if de:
                                lhs = replace(lhs, var, new_var)
                                equations[i][j] = (f"{lhs} = dt * ({rhs})", scope)

            if add_to_args:

                # save updated left-hand side variable to arguments dictionary
                for var_key, var in var_dict.copy().items():
                    if var_key == key or f"{var_key}_upd_" in key:
                        arg = var
                        break
                updated_args[f"{node}/{op}/{new_var}"] = arg

    return equations, updated_args


def update_equation_args(args: dict, updates: dict) -> dict:
    """Save variable updates to the equation args dictionary.

    Parameters
    ----------
    args
        Equation argument dictionary.
    updates
        Dictionary with variable updates.

    Returns
    -------
    dict
        Updated equation argument dictionary.
    """

    args_new = {}

    # add variables updates to equation arguments
    for key, arg in args.items():
        if key in updates:
            args_new[key] = updates[key]
        else:
            args_new[key] = arg

    # check which input fields need to be updated as well
    inputs = [key for key in args if 'inputs' in key]
    for inp in inputs:
        for in_key, in_map in args[inp].copy().items():
            for upd in updates:
                if in_map in upd:
                    args_new[inp].update({upd.split('/')[-1]: upd})
                    break

    return args_new


def parse_dict(var_dict: dict, backend, **kwargs) -> dict:
    """Parses a dictionary with variable information and creates backend variables from that information.

    Parameters
    ----------
    var_dict
        Contains key-value pairs for each variable that should be translated into the backend graph.
        Each value is a dictionary again containing the variable information (needs at least a field for `vtype`).
    backend
        Backend instance that the variables should be added to.
    kwargs
        Additional keyword arguments to be passed to the backend methods.

    Returns
    -------
    dict
        Key-value pairs with the backend variable names and handles.
    """

    var_dict_new = {}

    # go through dictionary items and instantiate variables
    #######################################################

    for var_name, var in var_dict.items():

        # instantiate variable
        if var['vtype'] == 'raw':
            var_dict_new[var_name] = var['value']
        else:
            var.update(kwargs)
            var_dict_new[var_name] = backend.add_var(name=var_name, **var)

    return var_dict_new


def split_equation(expr: str) -> tuple:
    """Splits an equation string into a left-hand side, right-and side and an assign type.

    Parameters
    ----------
    expr
        Equation string. Should contain a left-hand side and a right-hand side, separated by some form of assign symbol.

    Returns
    -------
    tuple
        left-hand side string, right-hand side string, assign operation string.
    """

    # define assign types and explicit non-assign types
    assign_types = ['+=', '-=', '*=', '/=']
    not_assign_types = ['<=', '>=', '==', '!=']

    lhs, rhs, assign_type, found_assign_type = "", "", "", False

    # look for assign types in expression
    for assign_type in assign_types:

        if assign_type in expr:

             # split expression via assign symbol
            if f' {assign_type} ' in expr:
                lhs, rhs = expr.split(f' {assign_type} ', maxsplit=1)
            elif f' {assign_type}' in expr:
                lhs, rhs = expr.split(f' {assign_type}', maxsplit=1)
            elif f'{assign_type} ' in expr:
                lhs, rhs = expr.split(f'{assign_type} ', maxsplit=1)
            else:
                lhs, rhs = expr.split(assign_type, maxsplit=1)
            found_assign_type = True
            break

        elif '=' in expr:

            # assume standard assign
            assign_type = '='
            assign = True

            # check if `=` symbol marks an assign operation or not
            for not_assign_type in not_assign_types:
                if not_assign_type in expr:
                    expr_tmp = expr.replace(not_assign_type, '')
                    if '=' not in expr_tmp:
                        assign = False

            if assign:

                # split equation via `=` symbol
                if f' = ' in expr:
                    lhs, rhs = expr.split(f' = ', maxsplit=1)
                elif f' {assign_type}' in expr:
                    lhs, rhs = expr.split(f' =', maxsplit=1)
                elif f'{assign_type} ' in expr:
                    lhs, rhs = expr.split(f'= ', maxsplit=1)
                else:
                    lhs, rhs = expr.split(f"=", maxsplit=1)
                found_assign_type = True
                break

    if not found_assign_type:
        return lhs, rhs, False
    return lhs, rhs, assign_type


def replace(eq: str, term: str, replacement: str, rhs_only: tp.Optional[bool] = False,
            lhs_only: tp.Optional[bool] = False) -> str:
    """Replaces a term in an equation with a replacement term (save replacement).

    Parameters
    ----------
    eq
        Equation that includes the term.
    term
        Term that should be replaced.
    replacement
        Replacement for all occurences of term.
    rhs_only
        If True, replacements will only be performed in right-hand side of the equation.
    lhs_only
        IF True, replacements will only be performed in left-hand side of the equation.

    Returns
    -------
    str
        The updated equation.
    """

    # define follow-up operations/signs that are allowed to follow directly after term in eq
    allowed_follow_ops = '+=*/^<>=!.%@[]():, '

    # replace every proper appearance of term in eq with replacement
    ################################################################

    eq_new = ""
    idx = eq.find(term)

    # go through all appearances of term in eq
    while idx != -1:

        # get idx of sign that follows after term
        idx_follow_op = idx+len(term)

        # if it is an allowed sign, replace term, else not
        replaced = False
        if (idx_follow_op < len(eq) and eq[idx_follow_op] in allowed_follow_ops) or idx_follow_op == len(eq):
            eq_part = eq[:idx]
            if (rhs_only and "=" in eq_part) or (lhs_only and "=" not in eq_part) or (not rhs_only and not lhs_only):
                eq_new += f"{eq_part} {replacement}"
                replaced = True
        if not replaced:
            eq_new += f"{eq[:idx_follow_op]}"

        # jump to next appearance of term in eq
        eq = eq[idx_follow_op:]
        idx = eq.find(term)

    # add rest of eq to new eq
    eq_new += eq

    return eq_new


def is_diff_eq(eq: str) -> bool:
    """Checks whether `eq` is a differential equation or not.

    Parameters
    ----------
    eq
        Equation string.

    Returns
    -------
    bool
        True, if `eq` is a differential equation.

    """

    lhs, rhs, _ = split_equation(eq)
    if "d/dt" in lhs:
        de = True
    elif "'" in lhs:
        de = True
    elif "dt_test" in replace(rhs, "dt", "dt_test"):
        de = True
    else:
        de = False

    return de


def is_coupled(eqs: list) -> bool:
    """Checks whether a list of equations defines a set of coupled equations, i.e. at least one left-hand side variable
    appears in the right-hand side of another equation.

    Parameters
    ----------
    eqs
        List of equation strings

    Returns
    -------
    bool
        True, if at least one left-hand side variable appears in the right-hand side of another equation.
    """

    # extract lhs, rhs and scope of each equation
    lhs_col, rhs_col, scope_col = [], [], []
    for eq, scope in eqs:
        lhs, rhs, _ = split_equation(eq)
        lhs_col.append(lhs)
        rhs_col.append(rhs)
        scope_col.append(scope)

    # check whether equations are coupled
    for lhs, lhs_scope in zip(lhs_col, scope_col):
        lhs = lhs.replace(" ", "")
        lhs = lhs.replace("d/dt", "")
        lhs = lhs.replace("*", "")
        lhs = lhs.replace("'", "")
        if "[" in lhs:
            l_idx = lhs.index("[")
            r_idx = lhs.index("]")
            lhs = lhs.replace(lhs[l_idx:r_idx+1], "")
        for rhs, rhs_scope in zip(rhs_col, scope_col):
            if "__test_string__" in replace(rhs, lhs, "__test_string__") and rhs_scope == lhs_scope:
                return True

    return False



from pyrates.utility import plot_timeseries, grid_search, plot_psd, plot_connectivity
import numpy as np
import matplotlib.pyplot as plt
from seaborn import cubehelix_palette

__author__ = "Richard Gast"
__status__ = "Development"


# parameters
dt = 1e-4
T = 5.
inp = 3. + np.random.randn(int(T/dt), 1) * 1.0
Cs = [1., 2., 4.]
ei_ratio = np.arange(0.5, 2., 0.1)[::-1]
l3l5_ratio = np.arange(0.5, 2., 0.1)
k_l3l5_e = np.zeros((int(len(ei_ratio) * len(l3l5_ratio))))
k_l3l5_i = np.zeros_like(k_l3l5_e)

fig, ax = plt.subplots(ncols=len(Cs), nrows=2, figsize=(20, 15), gridspec_kw={})
for idx, C in enumerate(Cs):

    k_l5l3_i = np.zeros_like(k_l3l5_e) + C
    n = 0
    for r1 in ei_ratio:
        for r2 in l3l5_ratio:
            k_l3l5_e[n] = C * r1 * r2
            k_l3l5_i[n] = C * r2
            n += 1

    params = {'k_l3l5_e': k_l3l5_e, 'k_l3l5_i': k_l3l5_i, 'k_l5l3_i': k_l5l3_i}
    param_map = {'k_l3l5_e': {'var': [(None, 'weight')],
                              'edges': [('L3/PC.0', 'L5/PC.0', 0)]},
                 'k_l3l5_i': {'var': [(None, 'weight')],
                              'edges': [('L3/PC.0', 'L5/IIN.0', 0)]},
                 'k_l5l3_i': {'var': [(None, 'weight')],
                              'edges': [('L5/PC.0', 'L3/IIN.0', 0)]},
                 }

    # perform simulation
    results = grid_search(circuit_template="EI_circuit.CMC",
                          param_grid=params, param_map=param_map,
                          inputs={("L3/PC.0", "Op_e.0", "i_in"): inp}, outputs={"r": ("L3/PC.0", "Op_e.0", "r")},
                          dt=dt, simulation_time=T, permute_grid=False, sampling_step_size=1e-3)

    # plotting
    cut_off = 1.
    max_freq = np.zeros((len(ei_ratio), len(l3l5_ratio)))
    freq_pow = np.zeros_like(max_freq)
    for k1, k2, k3 in zip(params['k_l3l5_e'], params['k_l3l5_i'], params['k_l5l3_i']):
        if not results[k1][k2][k3].isnull().any().any():
            _ = plot_psd(results[k1][k2][k3], tmin=cut_off, show=False)
            pow = plt.gca().get_lines()[-1].get_ydata()
            freqs = plt.gca().get_lines()[-1].get_xdata()
            r, c = np.argmin(np.abs(ei_ratio - k1/k2)), np.argmin(np.abs(l3l5_ratio - k2/k3))
            max_freq[r, c] = freqs[np.argmax(pow)]
            freq_pow[r, c] = np.max(pow)
            plt.close(plt.gcf())

    cm1 = cubehelix_palette(n_colors=int(len(ei_ratio)*len(l3l5_ratio)), as_cmap=True, start=2.5, rot=-0.1)
    cm2 = cubehelix_palette(n_colors=int(len(ei_ratio)*len(l3l5_ratio)), as_cmap=True, start=-2.0, rot=-0.1)
    cax1 = plot_connectivity(max_freq, ax=ax[0, idx], yticklabels=list(np.round(ei_ratio, decimals=2)),
                             xticklabels=list(np.round(l3l5_ratio, decimals=2)), cmap=cm1)
    cax1.set_xlabel('L3/L5 coupling strength')
    cax1.set_ylabel('exc/inh coupling strength')
    cax1.set_title(f'max freq (C = {C})')
    cax2 = plot_connectivity(freq_pow, ax=ax[1, idx], yticklabels=list(np.round(ei_ratio, decimals=2)),
                             xticklabels=list(np.round(l3l5_ratio, decimals=2)), cmap=cm2)
    cax2.set_xlabel('L3/L5 coupling strength')
    cax2.set_ylabel('exc/inh coupling strength')
    cax2.set_title(f'freq power (C = {C})')

plt.suptitle('Cortical microcircuit sensitivity to inter-laminar coupling')
plt.tight_layout(pad=2.5, rect=(0.01, 0.01, 0.99, 0.96))
fig.savefig("/home/rgast/Documents/Studium/PhD_Leipzig/Figures/BGTCS/cmc_coupling", format="svg")
plt.show()
